'''
 script to create
 - conecepts.py
 - __init__.py
 
 Bjoern Annighoefer 2023
'''

import os
import shutil

if __name__ == "__main__":
    GEN_FILE = "../cloner.py"
    GEN_TEMP = "cloner.py.mako"
    
    os.system("python -m eoq3conceptsgen.generatefromconceptscli -i %s -o %s"%(GEN_TEMP,GEN_FILE))


'''
A class managing watched expressions

2024 Bjoern Annighoefer
'''

from .qryrunner import QryRunner
from .change import WatChange

from ..value import VAL, U32, STR, NON, InitValOrNon
from ..query import Qry
from ..error import EOQ_ERROR_DOES_NOT_EXIST, EOQ_ERROR_RUNTIME, EoqErrorAsString
from ..config import Config
from ..serializer import CreateSerializer

from typing import Dict, Tuple, List


class WatchedQuery:
    def __init__(self, wid:int, queryStr:str, query:Qry):
        self.wid = wid
        self.queryStr = queryStr
        self.query = query
        self.watchers:Dict[str,bool] = {} #list of all observers. If the list is empty the query can be removed from the watchlist
        
        
class WatchEvaluationRecord:
    def __init__(self, wid:int):
        self.wid = wid #the id of the query to watch
        self.resultHash = None
        self.errorStr = None #the exception text
        self.evalCounter = 0 #how often was the constraint evaluated
        
class Watcher:
    '''Information on somebody observing a query
    '''
    def __init__(self,sessionId:str):
        self.sessionId = sessionId
        self.watchedQueries:Dict[int,WatchEvaluationRecord] = {}
        


class QueryWatcher:
    '''A class managing watched expressions
    '''
    def __init__(self,qryRunner:QryRunner,config:Config):
        self.qryRunner = qryRunner
        #internals
        self.expressionSerializer = CreateSerializer(config.constraintSerializer)
        self.currentWatchId = 0
        self.allWatchedQueries:Dict[int,WatchedQuery] = {} #watch ID for each unique query
        self.watchers:Dict[str,Watcher] = {} #each watcher is stored with the queries he watches and information about the latest evaluation
        self.superWatchers:Dict[str,Watcher] = {} #super watchers just observe every observed query as well
         
    def WatchQuery(self, queryStr:str, sessionId:str)->int:
        # see if the querry is already known or create a new one
        watchedQuery = next((x for x in self.allWatchedQueries.values() if x.queryStr == queryStr), None)
        if(None == watchedQuery): #create a new entry with new wid
            query = self.expressionSerializer.DesQry(queryStr) #use the same serializer as for constraints
            wid = self.currentWatchId
            watchedQuery = WatchedQuery(wid,queryStr,query)
            self.currentWatchId += 1
            self.allWatchedQueries[wid] = watchedQuery
            self.__AddWatchedQueryToSuperwatchers(wid)
        watchedQuery.watchers[sessionId] = True
        # add the query to the session
        self.__AddWatchedQueryToSession(watchedQuery.wid,sessionId)
        return watchedQuery.wid
    
    def WatchQueryById(self,wid:int,sessionId:str)->None:
        # look if a this wid has attached a query
        if(wid in self.allWatchedQueries):
            watchedQuery = self.allWatchedQueries[wid]
            watchedQuery.watchers[sessionId] = True
        else:
            raise EOQ_ERROR_DOES_NOT_EXIST('Unknown watch ID: %d'%(wid))
        # add the query to the session
        self.__AddWatchedQueryToSession(watchedQuery.wid,sessionId)
    
    def UnwatchQuery(self, queryStr:str, sessionId:str)->int:
        watchedQuery = next((x for x in self.allWatchedQueries.values() if x.queryStr == queryStr), None)
        if(watchedQuery):
            wid = watchedQuery.wid
            if(sessionId in watchedQuery.watchers):
                del watchedQuery.watchers[sessionId]
            self.__RemoveWatchedQueryIfNotWatchedAnyMore(watchedQuery)
            self.__RemoveWatchedQueryFromSession(wid,sessionId)
            return watchedQuery.wid
        else:
            raise EOQ_ERROR_DOES_NOT_EXIST('Unknown query: %s'%(queryStr))
                
    def UnwatchQueryById(self,wid:int,sessionId:str)->int:
        if(wid in self.allWatchedQueries):
            watchedQuery = self.allWatchedQueries[wid]
            if(sessionId in watchedQuery.watchers):
                del watchedQuery.watchers[sessionId]
            self.__RemoveWatchedQueryIfNotWatchedAnyMore(watchedQuery)
            self.__RemoveWatchedQueryFromSession(wid,sessionId)
            return wid
        else:
            raise EOQ_ERROR_DOES_NOT_EXIST('Unknown watch ID: %d'%(wid))
    
    def SuperWatch(self, sessionId:str)->None:
        if(not sessionId in self.superWatchers):
            superWatcher = Watcher(sessionId)
            #evaluate all queries for the superwatcher once
            for w in self.allWatchedQueries:
                superWatcher.watchedQueries[w.wid] = self.__InitEvalRecord(w.wid)
            #add the new superwatcher
            self.superWatchers[sessionId] = superWatcher
                
        
    def SuperUnwatch(self, sessionId:str)->None:
        if(sessionId in self.superWatchers):
            del self.superWatchers[sessionId]
            
    def GetAllWatcherSessionIds(self)->List[str]:
        allWatchingSessions = {w.sessionId:True for w in self.watchers.values()}
        for s in self.superWatchers.values():
            allWatchingSessions[s.sessionId] = True
        return list(allWatchingSessions.keys())
        
    def GetAllWatchChangesForSession(self, sessionId:str)->List[WatChange]:
        '''Evaluates all watched queries for a given sessionId
        WARNING: make sure that the corresponding session is set for the qryrunner,
        before this function is called
        '''
        changes = []
        if(sessionId in self.superWatchers):
            w = self.superWatchers[sessionId]
            changes = self.__GetAllChangesForWatcher(w)
        elif(sessionId in self.watchers): #only evaluate individual watches i not a superwatcher anyway. This prevents any duplicated evaluation
            w = self.watchers[sessionId]
            changes = self.__GetAllChangesForWatcher(w)
        return changes
    
    ### Internals functions
    def __AddWatchedQueryToSession(self,wid:int,sessionId:str)->None:
        watcher = None
        if(sessionId in self.watchers):
            watcher = self.watchers[sessionId]
        else:
            watcher = Watcher(sessionId)
            self.watchers[sessionId] = watcher
        if(not wid in watcher.watchedQueries):
            watcher.watchedQueries[wid] = self.__InitEvalRecord(wid)
            
    def __RemoveWatchedQueryFromSession(self,wid:int,sessionId:str):
        if(sessionId in self.watchers):
            watcher = self.watchers[sessionId]
            if(wid in watcher.watchedQueries):
                del watcher.watchedQueries[wid]
            
    def __EvalWatchedQuery(self, evaluationRecord:WatchEvaluationRecord)->Tuple[VAL,bool]:
        result = NON()
        changed = False
        watchedQuery = self.allWatchedQueries[evaluationRecord.wid] #this is not checked, because the function is only called internally
        try:
            result = self.qryRunner.Eval(watchedQuery.query, None)
            resultHash = result.__hash__()
            changed = (None != evaluationRecord.errorStr or resultHash != evaluationRecord.resultHash)
            evaluationRecord.resultHash = resultHash
            evaluationRecord.errorStr = None #remove the error string, such that the next error will be recognized as change
            evaluationRecord.evalCounter += 1
        except Exception as e:
            errorStr = EoqErrorAsString(EOQ_ERROR_RUNTIME("Evaluation failed: %s"%(str(e))))
            changed = (errorStr != evaluationRecord.errorStr)
            evaluationRecord.errorStr = errorStr
        return result, changed
    
    def __InitEvalRecord(self, wid:int)->WatchEvaluationRecord:
        ''' Initializes the Evaluation record
        WARNING: make sure this happens in the session belonging to the watcher!
        '''
        newRecord = WatchEvaluationRecord(wid)
        #the first time a new evaluation needs to be triggered
        self.__EvalWatchedQuery(newRecord)
        return newRecord
    
    def __RemoveWatchedQueryIfNotWatchedAnyMore(self, watchedQuery:WatchedQuery)->None:
        '''Removes a watchedQuery completely if no watcher watches it explicitly anymore
        '''
        if(0 == len(watchedQuery.watchers)):
            self.__RemoveDeletedWatchedQueryFromSuperwatchers(watchedQuery.wid)
            del self.allWatchedQueries[watchedQuery.wid]
    
    def __AddWatchedQueryToSuperwatchers(self, wid:int)->None:
        '''Removes a deleted watched query from all superwatchers
        A deleted watched query is one not watched explicitly by a non super watcher any more.
        '''
        for s in self.superWatchers.values():
            if(not wid in s.watchedQueries):
                s.watchedQueries[wid] = self.__InitEvalRecord(wid)
    
    def __RemoveDeletedWatchedQueryFromSuperwatchers(self, wid:int)->None:
        '''Removes a deleted watched query from all superwatchers
        A deleted watched query is one not watched explicitly by a non super watcher any more.
        '''
        for s in self.superWatchers.values():
            if(wid in s.watchedQueries):
                del s.watchedQueries[wid]
                
    def __GetAllChangesForWatcher(self, watcher:Watcher)->List[WatChange]:
        ''' Evaluates all queries for a single watcher and returns a WatchEvent for each changed result
        WARNING: make sure the qryrunners session is changed to the watcher beforehand.
        '''
        changes = []
        for wid,evaluationRecord in watcher.watchedQueries.items():
            result, changed = self.__EvalWatchedQuery(evaluationRecord)
            if(changed):
                change = WatChange(U32(wid),result,InitValOrNon(evaluationRecord.errorStr, STR))
                changes.append(change)
        return changes
        
            


"""
 2019 Bjoern Annighoefer
"""

from .remotedomainbasea import RemoteDomainBaseA

from ..config import Config
#type checking
from typing import Tuple

class SerialDomain(RemoteDomainBaseA):
    """A domain that supports processing serialized commands directly,
    which is often better performing in multi-computer environments
    """
    def __init__(self, config:Config):
        super().__init__(config)
        
    def SerRawDo(self, cmdStr:str, sessionId:str=None, serType:str=None, readOnly:bool=False)->Tuple[str,str]:
        """Direct processing of serialized commands
        Args:
        - cmdStr: The serialized command
        - sessionId: the session under which this command is processed
        - serType: [3 letter str] the name of the serializer

        Returns:
        - resStr: The serialized results
        - serType: [3 letter str] The serializer name used for the results
        """
        raise NotImplemented()

    def Close(self):
        pass #is to be called if the domain closes down
    
                    
    
        
        


'''
A class that starts a domain in a separate process.

2022 Bjoern Annighoefer
'''

from .multiprocessingdomainclient import MultiprocessingQueueDomainClient
from .multiprocessingdomainhost import MultiprocessingQueueDomainHost

from ..config import Config, EOQ_DEFAULT_CONFIG
from ..error import EOQ_ERROR_RUNTIME

from multiprocessing import Queue, Process
from threading import Thread
import queue

#type checking
from typing import Dict,Any

class PROCESS_ADMIN_COMMANDS:
    STOP = 'STOP'
    CONFIRM = 'CONFIRM'
    SEPERATOR = ' '

def DomainProcess(domainFactory:callable, domainFactoryArgs:Any, shallForwardSerializedCmds:bool, cmdTxQueue:Queue, cmdRxQueue:Queue, adminQueueTx:Queue, adminQueueRx:Queue, hostConfig:Config):
    #print("DomainProcess started")
    domain = domainFactory(domainFactoryArgs)
    server = MultiprocessingQueueDomainHost(cmdTxQueue,cmdRxQueue,domain,shallForwardSerializedCmds,hostConfig)
    #print("DomainProcess server ready.")
    #confirm that the domain in the process was created successfully
    adminQueueRx.put(PROCESS_ADMIN_COMMANDS.CONFIRM)
    #enter command loop
    shallRun = True
    while(shallRun):
        commandStr = adminQueueTx.get()
        commandFrags = commandStr.split(PROCESS_ADMIN_COMMANDS.SEPERATOR)
        command = commandFrags[0]
        if(PROCESS_ADMIN_COMMANDS.STOP == command):
            shallRun = False
            adminQueueRx.put(PROCESS_ADMIN_COMMANDS.CONFIRM)
    server.Stop()
    domain.Close()


class DomainToProcessWrapper(MultiprocessingQueueDomainClient):
    def __init__(self, domainFactory:callable, domainFactoryArgs:Any=None, shallForwardSerializedCmds:bool=False, config:Config=EOQ_DEFAULT_CONFIG):
        '''
        Args:
        domainFactory: a function that returns an instance of the domain
        '''
        super().__init__(Queue(), Queue(), config)
        self.adminQueueTx = Queue()
        self.adminQueueRx = Queue()
        
        if(config.processLessMode): #in debug mode no process is created, but the server executed in the same Process
            # in the following rx and tx queues for commands and events are mirrored to establish the communication
            self.clientProcess = Thread(name="Domain Worker (Thread)", target=DomainProcess, args=(domainFactory,domainFactoryArgs,shallForwardSerializedCmds,self.cmdRxQueue,self.cmdTxQueue,self.adminQueueTx,self.adminQueueRx,config,))
        else:
            self.clientProcess = Process(name="Domain Worker (Process)", target=DomainProcess, args=(domainFactory,domainFactoryArgs,shallForwardSerializedCmds,self.cmdRxQueue,self.cmdTxQueue,self.adminQueueTx,self.adminQueueRx,config,))
            self.clientProcess.daemon = True #quit this if the hosting process closes
        self.clientProcess.start()
        #wait on the confirmation that the server really started
        try:
            confirm = self.adminQueueRx.get(timeout=config.connectTimeout)
            if(PROCESS_ADMIN_COMMANDS.CONFIRM != confirm):
                raise EOQ_ERROR_RUNTIME('Failed to init domain process.')
        except queue.Empty:
            raise EOQ_ERROR_RUNTIME('Failed to init domain process.')
        
    #@Override
    def Close(self):
        self.__IssueAdminCommand(PROCESS_ADMIN_COMMANDS.STOP)
        self.clientProcess.join()
        super().Close()
            
        
    def __IssueAdminCommand(self,command:str,args:list=[]):
        commandStr = PROCESS_ADMIN_COMMANDS.SEPERATOR.join([command]+args)
        self.adminQueueTx.put(commandStr)
        confirm = self.adminQueueRx.get()
        if(PROCESS_ADMIN_COMMANDS.CONFIRM != confirm):
            raise EOQ_ERROR_RUNTIME('Communication with domain process failed.')
        
    


'''
 2022 Bjoern Annighoefer
'''

from .remotedomainclient import RemoteDomainClient
from .queuebasedframecom import QueueBasedFrameCom

from ..config import Config, EOQ_DEFAULT_CONFIG

from multiprocessing import Queue
from threading import Semaphore, Lock


'''
 CLIENT
 
'''

class DO_STRATEGY:
    SYNC = 0
    ASYNC = 1


class ResultWaitInfo:
    def __init__(self,commandId,doStrategy:int,semaphore=None,callback=None):
        self.commandId = commandId
        self.doStrategy = doStrategy
        self.semaphore = semaphore
        self.callback = callback
        self.res = None #this can be used to transfare back the result to the sync thread
    


class MultiprocessingQueueDomainClient(RemoteDomainClient):
    def __init__(self,cmdTxQueue:Queue, cmdRxQueue:Queue, config:Config=EOQ_DEFAULT_CONFIG, name:str="MultiprocessingQueueDomainClient"):
        super().__init__(config,name)
        self.cmdTxQueue = cmdTxQueue
        self.cmdRxQueue = cmdRxQueue
        self.comController = QueueBasedFrameCom(cmdTxQueue, cmdRxQueue, config, self.OnSerFrmReceived)  
        self.lock = Lock()
        self.resSignal = Semaphore(0)
        self.resStr = None
        self.frmId = 0
        self.frmIdLock = Lock()
        self.comController.Start() #last action to make sure not any notifies are called before
    
    #@Override    
    def Close(self):
        self.comController.Stop()
        
    #@Override
    def SendSerFrm(self,frmStr:str)->None:
        """Need to be implemented because it is called within RemoteDomainClient
        """
        self.comController.SendRaw(frmStr)



'''
2019 Bjoern Annighoefer
'''

from .logger import Logger,RegisterLogger,DEFAULT_LOGGER_LEVELS

class ConsoleLogger(Logger):
    """A logger printing to the console
    """
    def __init__(self,activeLevels:int=DEFAULT_LOGGER_LEVELS.L2_WARNING):
        super().__init__(activeLevels)

    #@Override         
    def _Log(self,levelName:str,msg:str):
        print("%s: %s"%(levelName,msg))

RegisterLogger("COL",ConsoleLogger,[])


from .cmdstream import CmdStreamA, CmdStream, RecorderCmdStream, ResNameGenStream, DependencyResolverStream, ObjToHisStream, ToDomainCmdStream, FromDomainCmdStream, EoqFileInStream, EoqFileOutStream, DomainWrapperStream
from .resnamegen import SimpleResNameGen, FromDomainResNameGen
from .validateeoqfile import ValidateEoqFile
from .loadeoqfile import LoadEoqFile
from .saveeoqfile import SaveEoqFile


'''
2019 Bjoern Annighoefer
'''

from .logger import Logger,RegisterLogger,DEFAULT_LOGGER_LEVELS

class CompoundLogger(Logger):
    """A logger forwarding the log messages to multiple child loggers.
    Use AddLogger and RemoveLogger to manage the list of loggers.
    """
    def __init__(self,activeLevels:int=DEFAULT_LOGGER_LEVELS.L2_WARNING):
        super().__init__(activeLevels)
        self.loggers = []

    #@Override
    def _Log(self,levelName:str,msg:str):
        for logger in self.loggers:
            logger.Log(levelName,msg)

    def AddLogger(self,logger:Logger):
        """Add a logger to the list of loggers
        """
        self.loggers.append(logger)

    def RemoveLogger(self,logger:Logger):
        """Remove a logger from the list of loggers
        """
        self.loggers.remove(logger)

RegisterLogger("CPL",CompoundLogger,[])


'''
Bjoern Annighoefer 2022
'''

### CONFIG CLASS ###

from typing import List, Dict, Any

class Config:
    ''' A settings struct for EOQ related classes
    '''
    def __init__(self):
        #access control
        self.superAdminName                 :str              = "superadmin" #there must always one super admin. This is how he is called
        self.defaultOwner                   :str              = "admin" # the owner of every element
        self.defaultGroup                   :str              = "users" # the default group for every unset element
        self.defaultPermission              :int              = 0xFF0 #create, delete, update and read for owner and group. None for nobody
        #remote processing
        self.remoteFrmTxSerializer          :str              = "TXT" #the default serializer used for sending remote frames
        self.remoteCmdTxSerializer          :str              = "JSO" #the default serializer used for sending remote commands
        self.remoteFrmRxSerializer          :str              = "TXT" #the serializer allowed for incoming frames. For security reasons, this should not be JSC or PYT
        self.remoteCmdRxSerializers         :List[str]        = ["TXT","JSO"] #the serializers allowed for incoming commands. All named here will be accepted. For security reasons, this should not be JSC or PYT
        self.frameVersion                   :str              = "3.0.0"
        self.threadLoopTimeout              :float            = 0.1   #[seconds] the timeout used a threaded loop waiting for events, e.g. event pull or push loops
        self.enableRawEvents                :bool             = False #do not serialize and deserialize events, which is faster, but does not allow to look inside
        self.connectTimeout                 :float            = 5.0   #[seconds] the timeout after that establishing a remote connection is assumed to be failed
        self.commandTimeout                 :float            = 10.0  #[seconds] the timeout after an established remote connection is assumed to be failed
        self.remoteRawLogging               :bool             = False # If true, remote communication is logged to files. Is only for Debug
        #eoq files
        self.fileSerializer                 :str              = "TXT" #the default serializer used for remote frames
        #command processing
        self.cmdMaxProcessingTime           :float            = None #[seconds] the maximum time a command can last before it is canceled (NOT IMPLEMENTED)
        self.cmdCmpMaxNesting               :int              = 10   #the maximum number levels of compound commands can be nested (NOT IMPLEMENTED)
        self.cmdMaxChanges                  :int              = 100  #the maximum number of changes stored in the domain
        self.cmdMaxMsgLength                :int              = 200  #the maximum number of characters in a message 
        self.cmdMaxMsgKeyLength             :int              = 15   #the maximum number of characters in a message key
        self.cmdMaxSubCommands              :int              = None #the maximum number of subcommands in a compound command (NOT IMPLEMENTED)
        self.qryMaxSegments                 :int              = None #the maximum number of segments (NOT IMPLEMENTED)
        self.resMaxElements                 :int              = None #the maximum number of elements in a result (NOT IMPLEMENTED)
        self.evtNotifyAsync                 :bool             = True #if true, event notification is handled in a thread separated from the transaction processing
        #constraitns
        self.constraintSerializer           :str              = "TXT" #the serializer for constraints
        #mdb
        self.strIdSeparator                 :str              = '__' #the separator for string IDs, e.g. package name<strIdSeparator>class name
        self.conceptsOnly                   :bool             = True #If True, mdb accepts only concept, no native elements or names
        #mpl
        self.mplMinSaveTimeout              :float            = 10.0 #[seconds] the time to wait form more changes, until the domain is persisted
        self.mplMaxSaveTimeout              :float            = 5*60.0 #[seconds] the time after a change after that persistence is enforced
        #statistics
        self.enableStatistics               :bool             = False # collect statistics/benchmark data during operation?
        #logging
        self.activeLogLevels                :int              = 0b000011 #the log levels that are active, i.e. warn and error
        self.logger                         :str              = "COL"  #the logger to use, e.g. COL = console logger, FIL = file logger, NOL = no logging, CFL = console and file logger
        self.loggerInitArgs                 :Dict[str,Any]    = {} #the arguments used to create the logger
        self.logSerializer                  :str              = "TXT" #the serializer for log entries
        #advanced debugging
        self.printExpectedExceptionTraces   :bool             = False #Shall traces of expected/caught exceptions be printed?
        self.printUnexpectedExceptionTraces :bool             = False #Shall traces of unexpected/uncaught exceptions be printed?
        self.processLessMode                :bool             = False #Threads instead of processes are used, which enables better debugging

### GLOBAL CONFIG ###
    
EOQ_DEFAULT_CONFIG = Config()
    


'''
 2019 Bjoern Annighoefer
'''

from .cmdrunner import CmdRunner   

from ..config import Config, EOQ_DEFAULT_CONFIG
from ..logger import GetLoggerInstance
from ..command import Cmd, Evt
from ..domain import Domain

from ..mdb import Mdb
from ..accesscontroller import AccessController, NoAccessController
 
from threading import Lock
#type checking
from typing import Callable, List, Any

class DomainWithMdb(Domain):
    def __init__(self, mdb:Mdb, accessController:AccessController=NoAccessController(), config:Config=EOQ_DEFAULT_CONFIG):
        super().__init__()
        self.mdb = mdb
        self.accessController = accessController
        self.config = config
        self.logger = GetLoggerInstance(config)
        #internals 
        self.threadSafeLock = Lock()
        self.cmdRunner = CmdRunner(mdb, accessController, config)
        #init the access controller        
        
        
    def RawDo(self, cmd:Cmd, sessionId=None, readOnly=False):
        self.threadSafeLock.acquire()
        #self.logger.PassivatableLog(LogLevels.INFO,lambda : "cmd: %s"%(self.serializer.serialize(cmd)))
        res = self.cmdRunner.Exec(cmd,sessionId,readOnly)
        self.threadSafeLock.release()
        return res
    
    #Override the event provide methods since the sole event provider shall be the cmd runner
    #@Override
    def Observe(self, callback:Callable[[List[Evt],Any,object],None], context:Any=None, sessionId:str=None)->None: #by default register for all events
        self.threadSafeLock.acquire()
        self.cmdRunner.Observe(callback,context,sessionId)
        self.threadSafeLock.release()
    
    #@Override    
    def Unobserve(self, callback:Callable[[List[Evt],Any,object],None], context:Any=None, sessionId:str=None)->None:
        self.threadSafeLock.acquire()
        self.cmdRunner.Unobserve(callback,context,sessionId)
        self.threadSafeLock.release()
        
    def Close(self):
        self.cmdRunner.Close()
        #benchmark:
        if self.config.enableStatistics:
            self.cmdRunner.benchmark.SaveToFile('CmdBenchmark.csv')
            self.logger.Info("Command benchmark saved to CmdBenchmark.csv")
            self.cmdRunner.qryRunner.benchmark.SaveToFile('QryBenchmark.csv')
            self.logger.Info("Query segment benchmark saved to QryBenchmark.csv")
        
        
 
        


'''
 2019 Bjoern Annighoefer
'''

'''
    Server
'''

from .remotedomainserver import RemoteDomainServer
from .queuebasedframecom import QueueBasedFrameCom

from ..config import Config
from ..domain import Domain

from multiprocessing import Queue


class MultiprocessingQueueDomainHost(RemoteDomainServer):
    def __init__(self,cmdTxQueue:Queue, cmdRxQueue:Queue, domain:Domain, shallForwardSerializedCmds:bool, config:Config):
        super().__init__(domain,shallForwardSerializedCmds,config)
        #internals
        self.shallRun = True
        self.comController = QueueBasedFrameCom(cmdTxQueue,cmdRxQueue,config,self.OnSerFrmReceived)
        self.comController.Start() #last action to make sure not any notifies are called before
        
    #@Override
    def SendSerFrm(self, frmStr:str, sessionId:str)->None:
        '''Needs to be overwritten by the implementation
        '''
        self.comController.SendRaw(frmStr)
           
    def Stop(self)->None:
        self.comController.Stop()
        
    
        


'''
 Bjoern Annighoefer 2019
'''

from .serializer import Serializer, RegisterSerializer

from ..value import VAL, BOL, U32, U64, I32, I64, F32, F64, STR, DAT, NON, QRY, LST
#import all queries because the eval function needs them
from ..query import Seg, Qry, Obj, Oni, His, Pth, Cls, Ino, Try, Met, Qrf, Stf, Slf, Not, Idx, Sel, Arr, Zip, Any, All, Equ, Eqa, Neq, Les, Gre, Rgx
#import all commands because the eval function needs them
from ..command import Cmd, Get, Upd, Del, Set, Add, Rem, Mov, Clo, Crt, Cpr, Mrg, Hel, Ses, Gby, Sts, Chg, Obs, Ubs, Msg, Ver, Gac, Res, Err, Evt, Scc, Cus, Cmp, GetCmdInfo
from ..frame import Frm
from ..error import EOQ_ERROR_INVALID_TYPE, EOQ_ERROR_INVALID_VALUE



class PySerializer(Serializer):
    '''Python Serializer   
    '''
    # @Overrides 
    def Name(self)->str:           
        return "PYT"
    # @Overrides 
    def SerFrm(self, frm:Frm)->str:
        return "Frm('%s',%d,'%s','%s','%s',%s,'%s')"%(frm.typ,frm.uid,frm.ser,frm.dat.replace("\\","\\\\").replace("'","\\'"),frm.sid,"True" if frm.roc else "False",frm.ver);
    
    # @Override
    def SerCmd(self, cmd:Cmd):
        return self.__SerCmdRaw(cmd,0)
    
    def __SerCmdRaw(self, cmd:Cmd, level:int):
        if(isinstance(cmd,Cmp)):
            cmdStrs = ['.' + self.__SerCmdRaw(c,level+1) for c in cmd.a]
            if(0==level):
                return "Cmp("+self.__CmdMuteResNameStr(cmd,"")+")"+"".join(cmdStrs)
            else:
                return "Append(Cmp("+self.__CmdMuteResNameStr(cmd,"")+")"+"".join(cmdStrs)+")"
        else:
            argStrs = [self.SerVal(a) for a in cmd.a]
            cmdInfo = GetCmdInfo(cmd.cmd)
            if(None != cmdInfo):
                if(cmdInfo.unpackArgs):
                    return "%s(%s%s)"%(self.__ToPyName(cmd.cmd),",".join(argStrs),self.__CmdMuteResNameStr(cmd,("," if 0<len(argStrs) else "")))
                else:
                    return "%s([%s]%s)"%(self.__ToPyName(cmd.cmd),",".join(argStrs),self.__CmdMuteResNameStr(cmd,","))
            else: #must be a custom command 
                return "Cus('%s',[%s]%s)"%(cmd.cmd,",".join(argStrs),self.__CmdMuteResNameStr(cmd,","))
        
    def __CmdMuteResNameStr(self, cmd:Cmd, prefix:str):
        segments = []
        if(cmd.m):
            segments.append("mute=True")
        if(cmd.r):
            segments.append("resName='%s'"%(cmd.r))
        if(0 < len(segments)):
            return prefix+",".join(segments)
        else:
            return ""
        
    # @Override    
    def SerQry(self, qry:Qry):
        segStrs = [self.SerSeg(s) for s in qry.v]
        return '.'.join(segStrs)
    
    # @Override
    def SerSeg(self, seg:Seg):
        argStrs = [self.SerVal(a) for a in seg.v]
        return self.__ToPyName(seg.seg)+ '(' + ','.join(argStrs) + ')'
    
    # @Override    
    def SerVal(self,val:VAL):
        if(isinstance(val,NON)): # NON must be first, because NON is also instance of many of the others
            return 'NON()'
        elif(isinstance(val,BOL)): 
            return 'BOL(True)' if val.v else 'BOL(False)'
        elif(isinstance(val,U32)): 
            return 'U32(%d)'%(val.v)
        elif(isinstance(val,U64)): 
            return 'U64(%d)'%(val.v)
        elif(isinstance(val,I32)): 
            return 'I32(%d)'%(val.v)
        elif(isinstance(val,I64)): 
            return 'I64(%d)'%(val.v)
        elif(isinstance(val,F32)): 
            return 'F32(%f)'%(val.v)
        elif(isinstance(val,F64)): 
            return 'F64(%f)'%(val.v)
        elif(isinstance(val,STR)): 
            return "STR('" + val.v.replace("\\","\\\\").replace("'","\\'") + "')"
        elif(isinstance(val,DAT)): 
            return "DAT(%d,%d,%d,%d,%d,%d,%f,%f)"%(val.year(),val.month(),val.day(),val.hour(),val.minute(),val.second(),val.microsec(),val.utcoff())
        elif(isinstance(val,QRY)):
            return self.SerQry(val)
        elif(isinstance(val,LST)): #list must be the after str, because str is also a list
            valStrs = [self.SerVal(v) for v in val.v] #cannot use Val() here, because that decomposes the full list
            return 'LST(['+','.join(valStrs)+'])'
        else:
            raise EOQ_ERROR_INVALID_TYPE("Unknown value type %s."%(type(val).__name__))
        
    def __ToPyName(self,typeStr:str)->str:
        return typeStr[0] + typeStr[1:].lower()

    def DesFrm(self, data)->Frm:
        from ..frame import Frm
        try:
            return eval(data)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Frm: %s"%(e))
    
    def DesCmd(self, data)->Cmd:
        try:
            return eval(data)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Cmd: %s"%(e))
    
    def DesQry(self, data)->Qry:
        try:
            return eval(data)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Qry: %s"%(e))
    
    def DesSeg(self, data)->Seg:
        try:
            return eval(data)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Seg: %s"%(e))
    
    def DesVal(self, data):
        try:
            return eval(data)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Val: %s"%(e))
        
### make this serializer known globally 
RegisterSerializer(PySerializer)


'''
 Model Persistence Layer (MPL) Interface

 Bjoern Annighoefer 2021
'''
from ..query import Qry, Obj
from ..domain import Domain


class Mpl:
    ''' Definition of a basic MPL interfaces
    
        Any EOQ Model Persistence Layer (MPL) should implement this interfaces
    '''
    
    def Load(self)->None:
        ''' Initializes the persistence locally
        
            Loads stored models and elements from the persistence store, e.g.
            the file system.
        '''
        
        pass
    
    def Connect(self,domain:Domain, target:Qry, featureName:str, sessionId:str, trackDomainChanges:bool=False)->Obj:
        ''' Connects and synchronizes with an domain
        
            Elements from the persistence are transfered into the domain or 
            vice versa. 
            The MPL will listens for model in the domain changes and persist those. 
            
            Args:
                domain: The domain which is the target for the domain
                target: The the element which will become the parent of the MPLs root
                featureName: The feature of the parent where the MPL root is added.
                sessionId: All commands will be executed with this session ID. If track 
                    changes is enabled, make sure that the user related with the session
                    has sufficient rights to observe all events.
                trackDomainChanges: If true the MPL monitors changes in the domain continously
                    and updates the local model accordingly
                    
            Returns:
                The workspace root element
        '''
        pass
    
    def Store(self)->None:
        ''' Persists the actual model state to the permanent store
        
            Forces the MPL to persist the current model to the permanent store, e.g. 
            the in memory content is written to the file system.          
            Store might also be carried out automatically by the MPL while changes occur. 
        '''
        pass
    
    def Close(self)->None:
        '''Gracefully shutting down and cleaning up the persistency
        '''
        pass
    
    
    


from .logger import Logger,\
    LOG_LEVEL_NAMES,\
    LOG_LEVELS_NUM,\
    DEFAULT_LOGGER_LEVELS,\
    RegisterLogger,\
    GetLoggerInstance
from .nologger import NoLogger
from .compoundlogger import CompoundLogger
from .consolelogger import ConsoleLogger
from .filelogger import FileLogger
from .consoleandfilelogger import ConsoleAndFileLogger



from .query import\
    SEG_TYPES,\
    MET_MODES,\
    IDX_MODES,\
    STF_MODES,\
    Seg,\
    Qry,\
    Obj,\
    Oni,\
    His,\
    Pth,\
    Cls,\
    Ino,\
    Try,\
    Met,\
    Qrf,\
    Stf,\
    Slf,\
    Not,\
    Idx,\
    Sel,\
    Arr,\
    Zip,\
    Any,\
    All,\
    Equ,\
    Eqa,\
    Neq,\
    Les,\
    Gre,\
    Rgx,\
    IsObj,\
    ObjId,\
    IsHis,\
    HisId


'''
2022 Bjoern Annighoefer

'''
from .accesscontroller import AccessController

class NoAccessController(AccessController):
    ''' Simple implementation that will do no access checking and no user management. 
    Everybody can do everything
    '''
    def __init__(self):
        super().__init__()
    
    ### INITIAL ACCESS ###
    
    def AuthenticateUser(self, user:str, password:str)->bool:
        return True
    
    ### OBSERVATION ###
    
    def IsAllowedToObserve(self, user:str, eventType:str)->bool:
        return True
    
    def IsAllowedToSuperobserve(self, user:str, eventType:str)->bool:
        return True



"""
 2025 Bjoern Annighoefer
"""

from ..domain import Domain
from ..config import Config
from ..command import Cmd
from ..error import EOQ_ERROR_ACCESS_DENIED
from ..logger import GetLoggerInstance
from ..serializer import CreateSerializer
from abc import ABC


class RemoteDomainBaseA(Domain, ABC):
    """Abstract base class for remote domains that are accessed via a network connection.
    Initialize the domain with the configuration and the logger and the serializers for incoming and outgoing data.
    """
    def __init__(self, config:Config):
        super().__init__()
        self.config = config
        self.logger = GetLoggerInstance(config)
        # create the serializers for outgoing data
        self.remoteFrmTxSerializer = CreateSerializer(config.remoteFrmTxSerializer)
        self.remoteCmdTxSerializer = CreateSerializer(config.remoteCmdTxSerializer)
        # create the serializers for incoming data
        self.remoteFrmRxSerializer = CreateSerializer(config.remoteFrmRxSerializer)
        self.remoteCmdRxSerializers = {s: CreateSerializer(s) for s in config.remoteCmdRxSerializers}

    def _DesCmd(self, serType:str, cmdStr:str) -> Cmd:
        """Deserialize a command
        Args:
        - serType: [3 letter str] the name of the serializer
        - cmdStr: The serialized command

        Returns:
        - cmd: The deserialized command
        """
        try:
            return self.remoteCmdRxSerializers[serType].DesCmd(cmdStr)
        except KeyError:
            raise EOQ_ERROR_ACCESS_DENIED("Serializer %s is not allowed for received commands" % (serType))

    def _SerCmd(self, serType:str, cmd:Cmd) -> str:
        """Serialize a command
        Args:
        - serType: [3 letter str] the name of the serializer
        - cmd: The command to serialize
        
        Returns:
        - cmdStr: The serialized command
        """
        try:
            return self.remoteCmdRxSerializers[serType].SerCmd(cmd)
        except KeyError:
            raise EOQ_ERROR_ACCESS_DENIED("Serializer %s is not allowed for sent commands" % (serType))







'''
 This file is generated form: gen/__init__.py.mako on 2024-07-11 15:05:08.891320 by calling E:/1_ILS/Projects/2018_EOQ/Groups/PYEOQ/pyeoq/eoq3conceptsgen/generatefromconceptscli.py -i gen/__init__.py.mako -o __init__.py.
 
 Bjoern Annighoefer 2023
'''

from .concepts import\
    CONCEPTS,\
    MXMDB,\
    MXELEMENT,\
    MXCONSTRAINT,\
    M2PRIMITIVES,\
    M2PACKAGE,\
    M2MODEL,\
    M2ENUM,\
    M2OPTIONOFENUM,\
    M2CLASS,\
    M2ATTRIBUTE,\
    M2ASSOCIATION,\
    M2COMPOSITION,\
    M2INHERITANCE,\
    M1MODEL,\
    M1OBJECT,\
    M1FEATURE,\
    M1ATTRIBUTE,\
    M1ASSOCIATION,\
    M1COMPOSITION,\
    CONCEPT_PREFIX,\
    CONCEPT_PREFIX_LEN,\
    CONCEPT_UNIQUE_LEN,\
    FEATURE_MULTIVALUE_POSTFIX,\
    IsConcept,\
    IsMultivalueFeature,\
    NormalizeFeatureName,\
    GetConceptKeyString



'''
Access controlled MDB wrapper. Stops if the user is not allowed to do the current action.

 Bjoern Annighoefer 2022
'''

from .session import Session

from ..mdb import Mdb
from ..error import EOQ_ERROR, EOQ_ERROR_UNKNOWN
from ..value import VAL, PRM, I64, STR, NON, LST
from ..query import Obj
from ..verifier import Verifier

#type checking
from typing import Tuple, Union


class VerifiedMdb(Mdb):
    def __init__(self, mdb:Mdb, verifier:Verifier):
        self.__mdb = mdb
        self.verifier = verifier
        self.session = None
        
    def SetSession(self, session:Session):
        '''Before any operation on the mdb, the current session must be set
        '''
        self.session = session
        
    def UnsetSession(self):
        '''Unset the session when the operation is over, to prevent any security breaches
        '''
        self.session = None
        
    def IsAllowedToRead(self, target:Obj, featureName:STR, user:str)->bool:
        ''' Checking read allowance without raising exceptions
        '''
        try:
            self.verifier.ReadPreVerify(target, featureName, NON(), user)
            return True
        except:
            return False
        
    #@Override
    def Create(self, classId:Union[STR,Obj], createArgs:LST=LST([]), target:Obj=NON(), recoveryArgs:LST=LST([]), shallVerify:bool=True, shallNotify=True)->Tuple[Obj,EOQ_ERROR]:
        #pre verification
        if shallVerify: self.verifier.CreatePreVerify(classId, createArgs, target, recoveryArgs, self.session.user)
        #creation
        (res,pmerr) = self.__mdb.Create(classId, createArgs, target, recoveryArgs)
        #post verification
        if(shallVerify and None == pmerr):
            #if operation was successful so far, look for carry out further checking and catch all post modification errors
            try:
                self.verifier.CreatePostVerify(classId, createArgs, res, recoveryArgs, self.session.user)
            except EOQ_ERROR as e:
                pmerr = e #store for later recovery
            except Exception as e:
                pmerr = EOQ_ERROR_UNKNOWN(str(e))
        #notification of successful creation
        if(shallNotify and None == pmerr):
            self.verifier.CreateNotify(classId, createArgs, res, self.session.user)
        return (res,pmerr)
    
    #@Override  
    def Read(self, target:Obj, featureName:STR, context:Obj=NON(), shallVerify:bool=True, shallNotify=True)->VAL:
        if shallVerify: self.verifier.ReadPreVerify(target,featureName,context,self.session.user)
        val = self.__mdb.Read(target,featureName,context)
        if shallVerify:
            val = self.__FilterReturnValue(val)
        #notification of successful read
        if(shallNotify):
            self.verifier.ReadNotify(target, featureName, context, val)
        return val
    
    #@Override
    def Update(self, target:Obj,featureName:STR, value:PRM, position:I64=I64(0), shallVerify:bool=True, shallNotify=True)->Tuple[Obj,Obj,Obj,I64]:
        #pre verification
        if(shallVerify):
            self.verifier.UpdatePreVerify(target, featureName, value, position, self.session.user)
        #actual update
        (oldValue, oldOwner, oldComposition, oldPosition, pmerr) = self.__mdb.Update(target, featureName, value, position)
        #post verification
        if(shallVerify and None == pmerr):
            #if operation was successful so far, look for carry out further checking and catch all post modification errors
            try:
                self.verifier.UpdatePostVerify(target, featureName, value, position, oldValue, self.session.user)
            except EOQ_ERROR as e:
                pmerr = e #store for later recovery
            except Exception as e:
                pmerr = EOQ_ERROR_UNKNOWN(str(e))
        #notification of a successful update
        if(shallNotify and None == pmerr):
            self.verifier.UpdateNotify(target, featureName, value, position, oldValue, self.session.user)
        return (oldValue, oldOwner, oldComposition, oldPosition, pmerr)
            
    
    #@Override
    def Delete(self, target:Obj, shallVerify:bool=True, shallNotify=True)->Tuple[STR,LST,LST]:
        if(shallVerify):
            self.verifier.DeletePreVerify(target, self.session.user)
        (classId, createArgs, recoveryArgs, pmerr) = self.__mdb.Delete(target)
        if(shallVerify and None == pmerr):
            #if operation was successful so far, look for carry out further checking and catch all post modification errors
            try:
                self.verifier.DeletePostVerify(target, classId, createArgs, recoveryArgs, self.session.user)
            except EOQ_ERROR as e:
                pmerr = e #store for later recovery
            except Exception as e:
                pmerr = EOQ_ERROR_UNKNOWN(str(e))
        #notify on successful delete
        if(shallNotify and None == pmerr):
            self.verifier.DeleteNotify(target, classId, createArgs, recoveryArgs, self.session.user)
        return (classId,createArgs,recoveryArgs,pmerr)
        
    #@Override
    def FindElementByIdOrName(self, classNameOrId:STR, context:Obj=NON(), restrictToConcept:STR=NON()) -> LST:
        return LST([c for c in self.__mdb.FindElementByIdOrName(classNameOrId,context,restrictToConcept) if self.IsAllowedToRead(c, NON(), self.session.user)])
    
    ### PRIVATE METHODS ###

    def __FilterReturnValue(self, val:VAL):
        '''Removes all not allowed objects from a list of return values
        '''
        if(LST == type(val)):
            return LST([v for v in val if (not isinstance(v,Obj) or self.IsAllowedToRead(v, NON(), self.session.user))])
        else:
            return val


'''
 2019 Bjoern Annighoefer
'''

from ..config import Config
from ..logger import GetLoggerInstance

from ..value import VAL, PRM, BOL, U32, U64, I64, STR, LST, NON, ValidateVal, ValidateValAndCast, InitValOrNon
from ..query import Qry, Obj
from ..command import CMD_TYPES, DEL_MODES, EVT_TYPES, EVT_KEYS, Cmd, Cmp, Err, Res, Evt, CrtEvt, UpdEvt, DelEvt, MsgEvt, VecEvt, WatEvt
from ..error import EOQ_ERROR, EOQ_ERROR_INVALID_VALUE, EOQ_ERROR_RUNTIME, EOQ_ERROR_UNSUPPORTED, EOQ_ERROR_ACCESS_DENIED, EOQ_ERROR_UNKNOWN, EOQ_ERROR_INVALID_OPERATION, EOQ_ERROR_CODES
from ..mdb import Mdb
from ..serializer import Serializer, TextSerializer, CreateSerializer
from ..util import Observable, GenerateSessionId
from ..concepts import CONCEPTS, MXELEMENT, M2CLASS, M2ATTRIBUTE, M2ASSOCIATION, M2COMPOSITION, M2INHERITANCE,\
                       M1MODEL, M1OBJECT, M1ATTRIBUTE, M1ASSOCIATION, M1COMPOSITION,\
                       IsConcept, NormalizeFeatureName 
from ..accesscontroller import AccessController
from ..util import Benchmark, SESSION_ID_LENGTH
from ..verifier import MultiVerifier

from .verifiedmdb import VerifiedMdb
from .constraintverifier import ConstraintVerifier
from .transaction import Transaction
from .session import Session
from .history import History
from .change import Change, CrtChange, UpdChange, DelChange, ObsChange, UbsChange, MsgChange, VecChange, WatChange
from .qryrunner import QryRunner
from .cloner import Cloner
from .deleter import Deleter
from .querywatcher import QueryWatcher
from .util import IsListOfObjects, IsList, IsNoList

from datetime import datetime
from threading import Thread, Event
import traceback

#type checking
from typing import Tuple, Dict, List
from pyecore.ecore import EModelElement


SESSION_SECRET_LEN = 10 #the part of the session id that is the secret

'''
 Cmd Runner
'''   
class CmdRunner(Observable):
    def __init__(self, mdb:Mdb, accessController:AccessController, config:Config):
        super().__init__()
        self.__mdb = mdb
        self.accessController = accessController
        self.config = config
        self.logger = GetLoggerInstance(config)
        self.logSerializer  = CreateSerializer(config.logSerializer)
        #initialize constraint verification
        self.constraintVerifier = ConstraintVerifier(self.config)
        #initialize the verification of the mdb
        self.verifier = MultiVerifier()
        self.verifier.Init(mdb)
        self.verifier.AddVerifier(self.accessController)
        self.verifier.AddVerifier(self.constraintVerifier)
        #wrap the mdb to make sure it is access verified
        self.mdb = VerifiedMdb(mdb,self.verifier)
        #internals
        self.cmdSerializer:Serializer = TextSerializer() #this is used for SCC
        #initialize internals
        self.latestTransactionId = 0
        self.sessionCount = 0
        self.transactions:Dict[int,Transaction] = {}
        #sessions
        self.sessions:Dict[str,Session] = {} #dict containing sessions keys and related session informations
        #changes
        self.earliestChangeId = 1
        self.latestChangeId = 1 # must start at 1 because otherwise -1 is created, which is an invalid change
        self.changes:Dict[int,Change] = {}
        #auxiliary modules
        self.qryRunner = QryRunner(self.mdb,self.config)
        self.deleter = Deleter(self,self.logger)
        self.cloner = Cloner(self,self.logger)
        self.queryWatcher = QueryWatcher(self.qryRunner,self.config) 
        #init command evaluators functor table
        self.cmdEvaluators = {}
        self.cmdEvaluators[CMD_TYPES.GET] = self.ExecGet
        self.cmdEvaluators[CMD_TYPES.UPD] = self.ExecUpd
        self.cmdEvaluators[CMD_TYPES.DEL] = self.ExecDel
        self.cmdEvaluators[CMD_TYPES.SET] = self.ExecSet
        self.cmdEvaluators[CMD_TYPES.ADD] = self.ExecAdd
        self.cmdEvaluators[CMD_TYPES.REM] = self.ExecRem
        self.cmdEvaluators[CMD_TYPES.MOV] = self.ExecMov
        self.cmdEvaluators[CMD_TYPES.CLO] = self.ExecClo
        self.cmdEvaluators[CMD_TYPES.CRT] = self.ExecCrt
        self.cmdEvaluators[CMD_TYPES.CPR] = self.ExecCpr
        self.cmdEvaluators[CMD_TYPES.MRG] = self.ExecMrg
        self.cmdEvaluators[CMD_TYPES.STS] = self.ExecSts
        self.cmdEvaluators[CMD_TYPES.HEL] = self.ExecHel
        self.cmdEvaluators[CMD_TYPES.SES] = self.ExecSes
        self.cmdEvaluators[CMD_TYPES.GBY] = self.ExecGby
        self.cmdEvaluators[CMD_TYPES.CHG] = self.ExecChg
        self.cmdEvaluators[CMD_TYPES.OBS] = self.ExecObs
        self.cmdEvaluators[CMD_TYPES.UBS] = self.ExecUbs
        self.cmdEvaluators[CMD_TYPES.MSG] = self.ExecMsg
        self.cmdEvaluators[CMD_TYPES.SCC] = self.ExecScc
        self.cmdEvaluators[CMD_TYPES.VER] = self.ExecVer
        self.cmdEvaluators[CMD_TYPES.GAC] = self.ExecGac
        self.cmdEvaluators[CMD_TYPES.CMP] = self.ExecCmp
        #init read only command (roc) functor evaluator table
        self.rocEvaluators = {}
        self.rocEvaluators[CMD_TYPES.GET] = self.ExecGet
        self.rocEvaluators[CMD_TYPES.CMP] = self.ExecCmp
        
        #custom commands
        self.customCommands = {}
    
        #create annonymous session
        anoSessionId = GenerateSessionId()
        (anoSessionIdPublic,anoSessionIdSecret) = self.__SplitSessionId(anoSessionId)
        self.annonymousSession = self.__CreateNewSession(anoSessionId,anoSessionIdPublic,anoSessionIdSecret)
            
        #benchmark
        if(self.config.enableStatistics):
            self.benchmark = Benchmark()
            allCMD_TYPES = [getattr(CMD_TYPES,k) for k in CMD_TYPES.__dict__ if not k.startswith('_')]
            for t in allCMD_TYPES:
                self.benchmark.InitMessure(t)
                
        if(self.config.evtNotifyAsync):
            #event notification thread
            self.notifyObserverSignal = Event()
            self.shallRun:bool = True
            self.evtNotificationThread:Thread = Thread(target=self.__EventNotificationLoop)
            self.evtNotificationThread.start()

        # merge / diff
        self.oldRoot = None
        self.newRoot = None
        self.postponedMergeReferences = []
        self.diffcommands = []
        self.mDummies = 1
        
    def Close(self):
        self.accessController.Deinit()
        if(self.config.evtNotifyAsync):
            self.shallRun = False
            self.evtNotificationThread.join()
        
    ### DIRECT MDB ACCESS
    # can also be called from the outside if the transaction id is known and valid
    
    def _MdbRead(self, tid:int, target:Obj,featureName:STR, context:Obj=NON()):
        res = self.mdb.Read(target,featureName,context)
        return res
    
    def _MdbCreate(self, tid:int, classId:STR, createArgs:LST=LST([]), target:Obj=NON(), recoveryArgs:LST=LST([])):
        (newElem,pmerr) = self.mdb.Create(classId,createArgs,target,recoveryArgs)
        isPostModError = None != pmerr
        change = CrtChange(newElem,classId,createArgs,isPostModError)
        self.__AddLocalChange(tid,change,True,True)
        if(isPostModError):
            raise pmerr
        return newElem
        
    def _MdbUpdate(self, tid:int, target:Obj, featureName:STR, value:PRM, position:I64=I64(0)):
        (oldVal,oldOwner,oldFeature,oldIndex,pmerr) = self.mdb.Update(target,featureName,value,position)
        isPostModError = None != pmerr
        change = UpdChange(target,featureName,value,position,oldVal,oldOwner,oldFeature,oldIndex,isPostModError) 
        self.__AddLocalChange(tid,change,True,True)
        if(isPostModError):
            raise pmerr
        
    def _MdbDelete(self, tid:int, target:Obj)->Tuple[STR,LST,LST]:
        (classId,createArgs,recoveryArgs,pmerr) = self.mdb.Delete(target)
        isPostModError = None != pmerr
        #create the del change itself
        change = DelChange(target,classId,createArgs,recoveryArgs,isPostModError) 
        self.__AddLocalChange(tid,change,True,True)
        #make sure deleted elements are not observed anymore. This prevents memory leaks. This must happen after the DelChange, because otherwise the change is not send
        self.__UnobserveDeletedElementForAllSessions(tid,target)
        if(isPostModError):
            raise pmerr
        return (classId,createArgs,recoveryArgs)
    
    def _MdbFindElementByNameOrId(self, tid:int, nameOrId:STR, context:Obj=NON(), restrictToConcept:STR=NON()) -> LST:
        res = self.mdb.FindElementByIdOrName(nameOrId, context, restrictToConcept)
        return res
        
    ### COMMAND HANDLERS
    
    def Exec(self, cmd:Cmd, sessionId:str=None, readOnly:bool=False) -> Cmd:
        res = None
        #determine the next transaction ID
        self.latestTransactionId+=1
        tid = self.latestTransactionId
        #run the transaction
        try:
            transaction = self.__StartTransaction(tid,sessionId,readOnly)
            val = self.ExecOnTransaction(cmd,tid,readOnly)
            history = self.__GetTransaction(tid).GetLastHistory()
            if(None!=history):
                (resNames,resNameIndicies) = history.GetValueNamesAndIndicies()
                res = Res(I64(tid),I64(transaction.cid),val,resNames,resNameIndicies)
            else:
                res = Res(I64(tid),I64(transaction.cid),val)
        except EOQ_ERROR as e:
            self.logger.Error("%d: %s"%(e.code,e.msg))
            if self.config.printExpectedExceptionTraces: 
                print("Expected exception trace (Exec):") 
                traceback.print_exc()
            res = Err(e.code,e.msg,e.trace)
        except Exception as e: 
            self.logger.Error("%d: %s"%(EOQ_ERROR_CODES.UNKNOWN,str(e)))
            if self.config.printUnexpectedExceptionTraces: 
                print("Unexpected exception trace (Exec):") 
                traceback.print_exc()
            res = Err(EOQ_ERROR_CODES.UNKNOWN,str(e))
        self.__EndTransaction(res,tid)
        return res
    
    def ExecOnTransaction(self, cmd:Cmd, tid:int, readOnly:bool) -> VAL:
        self.logger.PDebug(lambda : 'Transaction: %d: %s'%(tid,self.logSerializer.SerCmd(cmd)))
        cmdId = cmd.cmd
        val = NON()
        #1. find the routine to evaluate the comand type
        evaluator = None
        try:
            if(readOnly):
                evaluator = self.rocEvaluators[cmdId]
            else:
                evaluator = self.cmdEvaluators[cmdId]
        except KeyError: #it is no built-in command. Maybe a custom command?
            try:
                customCmd = self.customCommands[cmdId]
                evaluator = lambda a,t : self.ExecCus(customCmd,a,t)
            except KeyError:
                raise EOQ_ERROR_UNSUPPORTED("Unknown command type: %s."%(cmdId))
        #2. execute the command and catch potential errors
        if self.config.enableStatistics: self.benchmark.Start()
        try:
            val = evaluator(cmd.a,tid)
        except Exception as e: 
            if self.config.enableStatistics: self.benchmark.Stop(cmdId)
            raise e #forward exception
        if self.config.enableStatistics: self.benchmark.Stop(cmdId)
        #3. return the results
        self.logger.PDebug(lambda: 'Result: %d: %s'%(tid,self.logSerializer.SerVal(val)))
        return val
    
    def ExecCmp(self, args:list, tid:int) -> LST:
        transaction = self.__GetTransaction(tid)
        history = transaction.StartHistory()
        val = []
        n = 0;
        for cmd in args:
            n = n+1;
            try:
                v = self.ExecOnTransaction(cmd, tid, transaction.roc)
                history.AddValue(v,cmd.m,cmd.r)
                if(not cmd.m):
                    val.append(v)
            except EOQ_ERROR as e:
                transaction.StopHistory()
                e.msg = "Subcommand %d (%s): "%(n,cmd.cmd) + e.msg
                raise e
            except Exception as e:
                transaction.StopHistory()
                raise EOQ_ERROR_UNKNOWN("Subcommand %d (%s): %s"%(n,cmd.cmd,str(e)))
        transaction.StopHistory()
        return LST(val)
            
    def ExecGet(self,args,tid:int) -> VAL:
        target = ValidateVal(args[0],[VAL],False)
        history = self.__GetHistory(tid)
        val = self.qryRunner.Eval(target,history)
        return val
    
    def __GetConceptAndFeatureTypeForM1Object(self, tid:int, target:Obj, featureName:STR)->Tuple[STR,Obj]:
        '''Retrieves the concept type and feature element of an M1 object and given feature name
        '''
        normFeatureName = NormalizeFeatureName(featureName.GetVal())
        concept = self._MdbRead(tid, target, STR(MXELEMENT.CONCEPT), NON())
        if(CONCEPTS.M1OBJECT != concept.GetVal()):
            raise EOQ_ERROR_INVALID_OPERATION("%s is no M1OBJECT. Non-native features can only be resolved for M1OBJECTS."%(str(target)))
        clazz = self._MdbRead(tid, target, STR(M1OBJECT.M2CLASS), NON())
        features = self._MdbFindElementByNameOrId(tid,STR(normFeatureName),clazz,NON())
        nFeatures = len(features)
        if(0==nFeatures):
            raise EOQ_ERROR_INVALID_VALUE("%s is no known feature of %s."%(normFeatureName,str(target)))
        elif(1<nFeatures):
            raise EOQ_ERROR_INVALID_VALUE("%s is not unique for %s."%(normFeatureName,str(target))) #this should never happen
        feature = features[0]
        featureConcept = self._MdbRead(tid, feature, STR(MXELEMENT.CONCEPT), NON())
        return (featureConcept,feature)
    
    def __MultiElementModificationHandler(self, tid:int, target:VAL, feature:VAL, value:VAL, position:I64, updateWrapper) -> VAL:
        #set the value(s) depending on the multiplicity of the arguments
        if(IsNoList(target)): # e.g. #20
            if(IsNoList(feature)):
                if(IsNoList(value)):
                    updateWrapper(tid,target,feature,value,position)
                elif(IsList(value)):
                    for v in value:
                        updateWrapper(tid,target,feature,v,position)
                else:
                    raise EOQ_ERROR_INVALID_VALUE('Value must be single value or list of values, but got: %s:'%(value)) 
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for f in feature:
                        updateWrapper(tid,target,f,value,position)
                elif(IsListOfObjects(value) and len(value) == len(feature)):
                    for i in range(len(feature)):
                        if(IsNoList(value[i])):
                            updateWrapper(tid,target,feature[i],value[i],position)
                        elif(IsListOfObjects(value[i])):
                            for v in value[i]:
                                updateWrapper(tid,target,feature[i],v,position)
                        else:
                            raise EOQ_ERROR_INVALID_VALUE('For multiple features the value must be a list of list of objects for each feature, but entry %d is %s.'%(i,value[i]))
                else:
                    raise EOQ_ERROR_INVALID_VALUE('Value must be single value or list of list of values with outer list having a length equal to the number of features, but got: %s:'%(value)) 
            else:
                raise EOQ_ERROR_INVALID_VALUE('Feature must be single object or list of objects but got: %s:'%(feature)) 
        elif(IsListOfObjects(target)): # e.g. [#20,#22,#23]
            if(IsNoList(feature)):
                if(IsNoList(value)):
                    for t in target:
                        updateWrapper(tid,t,feature,value,position)
                elif(IsList(value) and len(value) == len(target)):
                    for j in range(len(target)):
                        if(IsListOfObjects(value[j])):
                            for i in range(len(value[j])):
                                updateWrapper(tid,target[j],feature,value[j][i],position)
                        else:
                            updateWrapper(tid,target[j],feature,value[j],position) #TODO is that OK
                            #raise EOQ_ERROR_INVALID_VALUE('For multiple targets the value must be a list of list of objects for each target, but entry %d is %s.'%(j,value[j]))
                else:
                    raise EOQ_ERROR_INVALID_VALUE('Value must be single value or list of values with equal length of the number of targets, but got: %s:'%(value)) 
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for t in target:
                        for f in feature:
                            updateWrapper(tid,t,f,value,position)
                elif(IsListOfObjects(value) and len(value) == len(feature)):
                    for t in target:
                        for i in range(len(feature)):
                            updateWrapper(tid,t,feature[i],value[i],position)
                elif(IsList(value) and len(value) == len(target)):
                    for j in range(len(target)):
                        if(IsListOfObjects(value[j]) and len(value[j]) == len(feature)):
                            for i in range(len(feature)):
                                updateWrapper(tid,target[j],feature[i],value[j][i],position)
                        elif(IsList(value[j]) and len(value[j]) == len(feature)):
                            for i in range(len(feature)):
                                if(IsList(value[j][i])):
                                    for v in value[j][i]:
                                        updateWrapper(tid,target[j],feature[i],v,position)
                                else:
                                    raise EOQ_ERROR_INVALID_VALUE('For multiple targets, multiple features and multiple values value must list equal to targets containing a list equal to features containing a list of values, but got %s for target %d and feature %d.'%(value[j][i],j,i)) 
                        else:
                            raise EOQ_ERROR_INVALID_VALUE('For multiple targets and multiple features the value for each entry must have the same length as the number of features. Expected %d entries for target %d, but got %s.'%(len(feature),j,value[j]))
                else:
                    raise EOQ_ERROR_INVALID_VALUE('Value must be single value or list of list of list of values with outer list having a length equal to the number of targets and the middle list with a length equal to the number of features, but got: %s:'%(value)) 
            else:
                raise EOQ_ERROR_INVALID_VALUE('Feature must be single object or list of objects but got: %s:'%(feature)) 
        else: 
            raise EOQ_ERROR_INVALID_VALUE('Target must be single object or list of objects but got: %s:'%(target))
    
    
    def __UpdUpdateWrapper(self, tid:int, target:Obj, featureName:STR, value:PRM, position:I64) -> None:
        self._MdbUpdate(tid, target, featureName, value, position)
        
    def __SetUpdateWrapper(self, tid:int, target:Obj, featureName:STR, value:PRM, position:I64) -> None:
        if(IsConcept(featureName.GetVal())):
            self._MdbUpdate(tid, target, featureName, value, position)
        else:
            (concept,feature) = self.__GetConceptAndFeatureTypeForM1Object(tid,target,featureName)
            if(CONCEPTS.M2ATTRIBUTE == concept):
                #check whether the attribute does already exist
                attrInstance = self._MdbRead(tid, target, STR(M1OBJECT.FEATUREINSTANCES+featureName.GetVal()))
                if(0<len(attrInstance)):
                    self._MdbUpdate(tid, attrInstance[0], STR(M1ATTRIBUTE.VALUE), value)
                else:
                    self._MdbCreate(tid, STR(CONCEPTS.M1ATTRIBUTE), LST([feature,target,value]))
            elif(CONCEPTS.M2ASSOCIATION == concept):
                self._MdbCreate(tid, STR(CONCEPTS.M1ASSOCIATION), LST([feature,target,value]))
            elif(CONCEPTS.M2COMPOSITION == concept):
                self._MdbCreate(tid, STR(CONCEPTS.M1COMPOSITION), LST([feature,target,value]))
            else:
                raise EOQ_ERROR_RUNTIME("Do not know how to add for feature %s for concept type %s."%(featureName,concept))
        
    def __AddUpdateWrapper(self, tid:int, target:Obj, featureName:STR, value:PRM, position:I64) -> None:
        if(IsConcept(featureName.GetVal())):
            self._MdbUpdate(tid, target, featureName, value, position)
        else:
            (concept,feature) = self.__GetConceptAndFeatureTypeForM1Object(tid,target,featureName)
            if(CONCEPTS.M2ATTRIBUTE == concept):
                self._MdbCreate(tid, STR(CONCEPTS.M1ATTRIBUTE), LST([feature,target,value]))
            elif(CONCEPTS.M2ASSOCIATION == concept):
                self._MdbCreate(tid, STR(CONCEPTS.M1ASSOCIATION), LST([feature,target,value]))
            elif(CONCEPTS.M2COMPOSITION == concept):
                self._MdbCreate(tid, STR(CONCEPTS.M1COMPOSITION), LST([feature,target,value]))
            else:
                raise EOQ_ERROR_RUNTIME("Do not know how to add for feature %s for concept type %s."%(featureName,concept))
        
    def __RemUpdateWrapper(self, tid:int, target:Obj, featureName:STR, value:PRM, position:I64) -> None:
        #position is never used, because it is determined from the value
        if(IsConcept(featureName.GetVal())):
            position = I64(self.mdb.Read(target,featureName).index(value))
            self._MdbUpdate(tid,target, featureName, NON(), position)
        else:
            (concept,feature) = self.__GetConceptAndFeatureTypeForM1Object(tid,target,featureName)
            self._MdbDelete(tid,feature)
    
    def ExecUpd(self, args:list, tid:int) -> LST:
        #eval all arguments
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[Obj,LST],'target')
        feature = ValidateVal(self.qryRunner.Eval(args[1],history),[STR,LST],'feature')
        value = ValidateVal(self.qryRunner.Eval(args[2],history),[VAL],'value',False)
        position = ValidateValAndCast(self.qryRunner.Eval(args[3],history),[I64],'position')
        # delegate to multi element modification
        self.__MultiElementModificationHandler(tid, target, feature, value, position, self.__UpdUpdateWrapper)
        val = LST([target,feature,position,value])
        return val
    
   
    
    def _Delete(self, target:Obj, mode:STR, tid:int)->Tuple[STR,LST,LST]:
        res = None
        if(DEL_MODES.BAS == mode):
            res = self._MdbDelete(tid,target)
        elif(DEL_MODES.AUT == mode):
            res = self.deleter.DeleteAuto(target, tid)
        elif(DEL_MODES.FUL == mode):
            res = self.deleter.DeleteFull(target, tid)
        else:
            EOQ_ERROR_INVALID_VALUE('Mode %s is not valid:'%(mode))
        return res
            
            
    def ExecDel(self, args:list, tid:int) -> LST:
        #eval all arguments
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[Obj,LST],'target')
        mode = ValidateValAndCast(self.qryRunner.Eval(args[1],history),[STR],'mode')
        res = LST([])
        if(IsListOfObjects(target)):
            for t in target:
                (conceptId,createArgs,recoveryArgs) = self._Delete(t, mode, tid)
                res.append(LST([conceptId,createArgs,recoveryArgs]))
        elif(isinstance(target,Qry)):
            (conceptId,createArgs,recoveryArgs) = self._Delete(target, mode, tid)
            res = LST([conceptId,createArgs,recoveryArgs])
        else:
            raise EOQ_ERROR_INVALID_VALUE('Target must be single object or list of objects but got: %s:'%(target))
        
        return res    
    
    
    def ExecSet(self, args:list, tid:int) -> LST:
        #eval all arguments
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[Obj,LST],'target')
        feature = ValidateVal(self.qryRunner.Eval(args[1],history),[STR,LST],'feature')
        value = self.qryRunner.Eval(args[2],history)
        # delelgate to multi element modification
        self.__MultiElementModificationHandler(tid, target, feature, value, I64(0), self.__SetUpdateWrapper)
        val = LST([target,feature,value])
        return val
    

    def ExecAdd(self, args:list, tid:int) -> LST:
        #eval all arguments
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[Obj,LST],'target')
        feature = ValidateVal(self.qryRunner.Eval(args[1],history),[STR,LST],'feature')
        value = self.qryRunner.Eval(args[2],history)
        # delelgate to multi element modification
        self.__MultiElementModificationHandler(tid, target, feature, value, I64(-1), self.__AddUpdateWrapper)
        val = LST([target,feature,value])
        return val
    
    def ExecRem(self, args:list, tid:int) -> LST:
        # DEPRICATED
        self.__DeprecatedWarning("REM")
        #eval all arguments
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[Obj,LST],'target')
        feature = ValidateVal(self.qryRunner.Eval(args[1],history),[STR,LST],'feature')
        value = ValidateVal(self.qryRunner.Eval(args[2],history),[VAL],'value',False)
        # delelgate to multi element modification
        self.__MultiElementModificationHandler(tid, target, feature, value, NON(), self.__RemUpdateWrapper)
        val = LST([target,feature,value])
        return val
    
    def __MovUpdateWrapper(self, target:Obj, feature:STR, value:PRM, position:U64 ,tid:int) -> None:
        # TODO
        parent = self.mdb.Read(value,STR(M1OBJECT.PARENT))
        featureName = self.mdb.Read(parent[1],STR(M2COMPOSITION.NAME))
        self._MdbUpdate(tid,parent[0], featureName, value, -(I64(position)+I64(2)))
    
    def ExecMov(self, args:list, tid:int) -> LST:
        # DEPRICATED
        self.__DeprecatedWarning("MOV")
        #eval all arguments
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[Obj,LST],'target')
        newIndex = ValidateValAndCast(self.qryRunner.Eval(args[1],history),[U64],'newIndex')
        
        #set the value(s) depending on the multiplicity of the arguments
        if(IsNoList(target)): # e.g. #20
            self.__MovUpdateWrapper(NON(),NON(),target,newIndex,tid)
        elif(IsListOfObjects(target)): # e.g. [#20,#22,#23]
            for t in target:
                self.__MovUpdateWrapper(NON(),NON(),t,newIndex,tid)
        else:
            raise EOQ_ERROR_INVALID_VALUE('Target must Obj or LST of Obj but got: %s:'%(target))
        
        val = LST([target,newIndex])
        return val
    
    
    
    def ExecClo(self, args:list, tid:int) -> VAL:
        #res = None  
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[Obj],'target')
        mode = ValidateValAndCast(self.qryRunner.Eval(args[1],history),[STR],'mode')
        createArgOverrides = ValidateValAndCast(self.qryRunner.Eval(args[2],history),[LST],'createArgOverrides')
        val = self.cloner.Clone(target,mode,createArgOverrides,tid)
        return val

    def ExecMrg(self, args, tid):
        '''
        Merges all changes from file 'new' onto file 'old'.
        '''
        pass
#         val = self.ExecCpr(args, tid)
#         cmds = res.GetValue()
#         # execute commands
#         for cmd in cmds:
#             self.ExecOnTransaction(cmd, tid)
#         return Res(CMD_TYPES.MRG, status, res, tid, self.latestChangeId)

    def ExecCpr(self, args, tid):
        pass
#         '''
#         Creates a list with all the necessary commands to adapt 
#         the old object to the new one.
#         '''
#         status = ResTypes.OKY
#         res = None
#         history = self.__GetHistory(tid)
#         old     = self.qryRunner.Eval(args[0], history) # old file
#         new     = self.qryRunner.Eval(args[1], history) # new file
#         mrgMode = self.qryRunner.Eval(args[2], history) # new file
#         if IsList(old) or IsList(new):
#             raise EOQ_ERROR(0,'Error in Diff/Merge: No Lists allowed.')
#         # init diff.py
#         differenceFinder = DifferenceFinder(self, self.mdb, 
#             old, new, tid, mrgMode, logger = self.logger)
#         if mrgMode == MrgModes.DET:
#             differenceFinder.CreateObjectOverviewFile()
#         # start diff.py
#         differenceFinder.Run()
#         # obtain commands
#         res = differenceFinder.GetCommands()
#         # serialize commands
#         if mrgMode in [MrgModes.LST, MrgModes.LOG, MrgModes.DET]:
#             path = 'log/cmdList_{}_{}.txt'.format(old, new)
#             with open(path, 'w') as file:
#                 file.write('### Commands to merge {} onto {} ###\n'.format(new, old))
#                 for cmd in res:
#                     file.write(self.logSerializer.Ser(cmd))
#                     file.write('\n')
#         return Res(CMD_TYPES.CPR, status, res, tid, self.latestChangeId)
    
    
    def ExecCrt(self, args:list, tid:int) -> VAL:
        res = NON()  
        history = self.__GetHistory(tid)
        classId = ValidateVal(self.qryRunner.Eval(args[0],history),[STR,Obj],'classId')
        n = ValidateValAndCast(self.qryRunner.Eval(args[1],history),[U32],"n").GetVal()
        createArgs = ValidateValAndCast(self.qryRunner.Eval(args[2],history),[NON,LST],"createArgs")
        target = ValidateValAndCast(self.qryRunner.Eval(args[3],history),[NON,Obj,LST],"target")
        recoveryArgs = ValidateValAndCast(self.qryRunner.Eval(args[4],history),[NON,LST],"createArgs")
        if n == 1:
            res = self._MdbCreate(tid,classId, createArgs, target, recoveryArgs)
        elif(IsListOfObjects(createArgs)):
            res = LST()
            for i in range(n):
                elem = self._MdbCreate(tid, classId, createArgs) #cannot use recovery here
                res.append(elem)
        elif(LST==type(createArgs) and n==len(createArgs)): #n > U32(1)
            res = LST()
            for i in range(n):
                elem = self._MdbCreate(tid, classId, createArgs[i]) #cannot use recovery here
                res.append(elem)
        elif(LST==type(createArgs) and 0==len(createArgs)): #n > U32(1)
            res = LST()
            for i in range(n):
                elem = self._MdbCreate(tid, classId) #cannot use recovery here
                res.append(elem)
        else:
            raise EOQ_ERROR_INVALID_VALUE("There must be exactly as many individual createArgs arrays as the number of elements to be created.")
        return res
    
    def ExecHel(self, args:list, tid:int) -> STR:
        transaction = self.__GetTransaction(tid)
        user = ValidateValAndCast(args[0],[STR],'user').GetVal()
        password = ValidateValAndCast(args[1],[STR],'password').GetVal()
        #prepare a new session
        #sessionId = None
        #check identification
        if(self.annonymousSession == transaction.session):
            raise EOQ_ERROR_ACCESS_DENIED('Cannot login on anonymous session. Please provide session ID.')
        if(self.accessController.AuthenticateUser(user, password)):
            session = transaction.session
            session.user = user
        else:
            raise EOQ_ERROR_ACCESS_DENIED('Identification failed.')
        val = BOL(True)
        return val
    
    def ExecSes(self, args:list, tid:int):
        transaction = self.__GetTransaction(tid)
        sessionId = ValidateValAndCast(args[0],[STR],'sessionId').GetVal() 
        (sessionIdPublic,sessionIdSecret) = self.__SplitSessionId(sessionId)
        try:
            session = self.sessions[sessionIdPublic]
            if(sessionIdSecret != session.sessionIdSecret):
                raise EOQ_ERROR_INVALID_VALUE('Unknown session %s'%(sessionId))
            transaction.session = session
            self.mdb.SetSession(session)
        except KeyError:
            raise EOQ_ERROR_INVALID_VALUE('Unknown session %s'%(sessionId))
        val = BOL(True)
        return val
    
    def ExecGby(self, args:list, tid:int) -> BOL:
        transaction = self.__GetTransaction(tid)
        session = transaction.session
        if(self.annonymousSession == session):
            raise EOQ_ERROR_ACCESS_DENIED("Unknown session.")
        session = self.sessions[session.sessionIdPublic]
        self.sessions.pop(session.sessionIdPublic) #delete the session
        transaction.session = None #remove any eventual remaining reference to the session
        val = BOL(True)
        return val
    
    def ExecSts(self, args:list, tid:int) -> U64:
        val = U64(self.latestChangeId-1) #latest the latest change is one below the indication
        return val
    
    def ExecChg(self, args:list, tid:int) -> LST:
        history = self.__GetHistory(tid)
        changeId = ValidateValAndCast(self.qryRunner.Eval(args[0],history),[U64],'changeId').GetVal()
        n = ValidateValAndCast(self.qryRunner.Eval(args[1],history),[U64],'n').GetVal()
        #sanity check
        if(n<0):
            raise EOQ_ERROR_INVALID_VALUE("Index must be greater or equal zero, but got: %d"%(n))
        val = NON()
        if(changeId<self.earliestChangeId):
            if(n==0 or changeId+n >= self.latestChangeId): 
                val = LST([self.__ChangeToChangeRecord(self.changes[i]) for i in range(self.earliestChangeId,self.latestChangeId)])
            elif(changeId+n >= self.earliestChangeId):
                val = LST([self.__ChangeToChangeRecord(self.changes[i]) for i in range(self.earliestChangeId,changeId+n)])
            else: 
                val = LST()
        elif(changeId >= self.latestChangeId):
            val = LST()
        else: #is a valid change id
            if(n==0 or changeId+n >= self.latestChangeId):
                val = LST([self.__ChangeToChangeRecord(self.changes[i]) for i in range(changeId,self.latestChangeId)])
            else:
                val = LST([self.__ChangeToChangeRecord(self.changes[i]) for i in range(changeId,changeId+n)])
        return val
    
    def ExecObs(self, args:list, tid:int) -> BOL: 
        transaction = self.__GetTransaction(tid)
        history = transaction.GetCurrentHistory()
        evtType = ValidateValAndCast(self.qryRunner.Eval(args[0],history),[STR],'evtType').GetVal() #everything must be string
        evtKeyUnchecked = self.qryRunner.Eval(args[1],history)
        #check is session is known
        if(not transaction.session):
            raise EOQ_ERROR_RUNTIME('Must specify session when observing events.')
        session = transaction.session
        (evtKeyChecked, val, isChanged) = self.__Obs(evtType, evtKeyUnchecked, history, session)
        #add to changes for making this undoable
        if(isChanged):
            self.__AddLocalChange(tid, ObsChange(evtType,evtKeyChecked),False,False)
        #return always true 
        return val
    
    def __Obs(self, evtType:str, evtKeyUnchecked:VAL, history:History, session:Session)->(VAL,VAL,bool):
        val = BOL(True)
        evtKey = evtKeyUnchecked
        isChanged = False
        #decide based on the event type what to do
        if(EVT_TYPES.ELM == evtType):
            evtKey = ValidateValAndCast(self.qryRunner.Eval(evtKeyUnchecked,history),[LST,Obj],'evtKey')
            if(isinstance(evtKey,Obj)):
                isChanged = session.ObserveElement(evtKey)
            elif(LST == type(evtKey)):
                if(IsListOfObjects(evtKey)):
                    isChanged, evtKey = session.ObserveElements(evtKey)
                else:
                    raise EOQ_ERROR_INVALID_VALUE("Multiple elements to be observed must be given as a flat list of objects.")
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        elif(EVT_TYPES.CRT == evtType or\
             EVT_TYPES.UPD == evtType or\
             EVT_TYPES.DEL == evtType or\
             EVT_TYPES.VEC == evtType):
            evtKey = ValidateValAndCast(self.qryRunner.Eval(evtKeyUnchecked,history),[STR,NON],'evtKey')
            if(evtKey.IsNone()):
                isChanged = session.ObserveEvent(evtType)
            elif(STR == type(evtKey) and EVT_KEYS.ALL == evtKey.GetVal()):
                if(self.accessController.IsAllowedToSuperobserve(session.user, evtType)):
                    isChanged = session.SuperobserveEvent(evtType)
                else:
                    raise EOQ_ERROR_ACCESS_DENIED('Not allowed to superobserve %s'%(evtType))
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        elif(EVT_TYPES.WAT == evtType):
            evtKey = ValidateValAndCast(evtKeyUnchecked,[STR,U32,NON],'evtKey')
            if(evtKey.IsNone()): #enable observation
                isChanged = session.ObserveEvent(evtType)
            elif(U32 == type(evtKey)):
                wid = evtKey.GetVal()
                self.queryWatcher.WatchQueryById(wid, session.sessionIdPublic)
                isChanged = session.ObserveQuery(wid)
            elif(STR == type(evtKey)):
                if(EVT_KEYS.ALL == evtKey.GetVal()): #enable superobservation
                    if(self.accessController.IsAllowedToSuperobserve(session.user, evtType)):
                        self.queryWatcher.SuperWatch(session.sessionIdPublic)
                        isChanged = session.SuperobserveEvent(evtType)
                    else:
                        raise EOQ_ERROR_ACCESS_DENIED('Not allowed to superobserve %s'%(evtType))
                else: #watch a normal query
                    if(self.accessController.IsAllowedToObserve(session.user, evtType)):
                        wid = self.queryWatcher.WatchQuery(evtKey.GetVal(), session.sessionIdPublic)
                        isChanged = session.ObserveQuery(wid)
                        val = U32(wid) #return watch id
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        elif(EVT_TYPES.MSG == evtType):
            evtKey = ValidateValAndCast(self.qryRunner.Eval(evtKeyUnchecked,history),[STR,NON],'evtKey')
            if(evtKey.IsNone()): #enable MSG observation
                isChanged = session.ObserveEvent(evtType)
            elif(STR == type(evtKey) and EVT_KEYS.ALL == evtKey.GetVal()): #enable super observation
                if(self.accessController.IsAllowedToSuperobserve(session.user, evtType)):
                    isChanged = session.SuperobserveEvent(evtType)
                else:
                    raise EOQ_ERROR_ACCESS_DENIED('Not allowed to superobserve %s'%(evtType))
            elif(STR == type(evtKey)): #set the event key and observe
                isChanged = session.ObserveMessage(evtKey)
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        else:
            raise EOQ_ERROR_UNSUPPORTED('Unknown event type: %s'%(evtType)) 
        return (evtKey,val,isChanged)
    
    def ExecUbs(self, args:list, tid:int) -> BOL: 
        transaction = self.__GetTransaction(tid)
        history = transaction.GetCurrentHistory()
        evtType = ValidateValAndCast(self.qryRunner.Eval(args[0],history),[STR],'evtType').GetVal() #everything must be string
        evtKeyUnchecked = args[1]
        #check is session is known
        if(not transaction.session):
            raise EOQ_ERROR_RUNTIME('Must specify session when unobserving events.')
        session = transaction.session
        (evtKey,val,isChanged) = self.__Ubs(evtType, evtKeyUnchecked, history, session)
        #add to changes for making this undoable
        if(isChanged):
            self.__AddLocalChange(tid, UbsChange(evtType,evtKey,session.sessionIdPublic),False,False)
        return val
    
    def __Ubs(self, evtType:str, evtKeyUnchecked:VAL, history:History, session:Session)->(VAL,VAL,bool):
        val = BOL(True)
        evtKey = evtKeyUnchecked
        isChanged = False
        #decide based on the event type what to do
        if(EVT_TYPES.ELM == evtType):
            evtKey = ValidateValAndCast(self.qryRunner.Eval(evtKeyUnchecked,history),[LST,NON,Obj],'evtKey')  #Obj must be last in row, because it has no construction check
            if(isinstance(evtKey,Obj)):
                isChanged = session.UnobserveElement(evtKey)
            elif(LST == type(evtKey)):
                if(IsListOfObjects(evtKey)):
                    isChanged, evtKey = session.UnobserveElements(evtKey)
                else:
                    raise EOQ_ERROR_INVALID_VALUE("Multiple objects to be observed must be given as a flat list of objects.")
            elif(evtKey.IsNone()):
                isChanged, evtKey = session.UnobserveAllElements()
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        elif(EVT_TYPES.CRT == evtType or\
            EVT_TYPES.UPD == evtType or\
             EVT_TYPES.DEL == evtType or\
             EVT_TYPES.VEC == evtType):
            evtKey = ValidateValAndCast(self.qryRunner.Eval(evtKeyUnchecked,history),[STR,NON],'evtKey')
            if(evtKey.IsNone()):
                isChanged = session.UnobserveEvent(evtType)
            elif(STR == type(evtKey) and EVT_KEYS.ALL == evtKey.GetVal()):
                isChanged = session.UnSuperobserveEvent(evtType)
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        elif(EVT_TYPES.WAT == evtType):
            evtKey = ValidateValAndCast(evtKeyUnchecked,[STR,U32,NON],'evtKey')
            if(evtKey.IsNone()): #enable observation
                isChanged = session.UnobserveEvent(evtType)
            elif(U32 == type(evtKey)):
                wid = evtKey.GetVal()
                self.queryWatcher.UnwatchQueryById(wid,session.sessionIdPublic)
                isChanged = session.UnobserveQuery(wid)
            elif(STR == type(evtKey)):
                if(EVT_KEYS.ALL == evtKey.GetVal()): #enable superobservation
                    self.queryWatcher.SuperUnwatch(session.sessionIdPublic)
                    isChanged = session.UnSuperobserveEvent(evtType)
                else:
                    wid = self.queryWatcher.UnwatchQuery(evtKey.GetVal(),session.sessionIdPublic)
                    if(0 <= wid): #otherwise the query was never watched
                        isChanged = session.UnobserveQuery(wid)
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        elif(EVT_TYPES.MSG == evtType):
            evtKey = ValidateValAndCast(self.qryRunner.Eval(evtKeyUnchecked,history),[STR,NON],'evtKey')
            if(evtKey.IsNone()): #enable MSG observation
                isChanged = session.UnobserveEvent(evtType)
            elif(STR == type(evtKey) and EVT_KEYS.ALL == evtKey.GetVal()): #enable super observation
                isChanged = session.UnSuperobserveEvent(evtType)
            elif(STR == type(evtKey)): #set the event key and observe
                isChanged = session.UnobserveMessage(evtKey)
            else:
                raise EOQ_ERROR_INVALID_VALUE("Unexpected event key: %s"%(evtKey))
        else:
            raise EOQ_ERROR_UNSUPPORTED('Unknown event type: %s'%(evtType)) 
        return (evtKey,val,isChanged)
    
    def ExecMsg(self, args:list, tid:int) -> BOL: 
        transaction = self.__GetTransaction(tid)
        history = transaction.GetCurrentHistory()
        msgKey = ValidateValAndCast(self.qryRunner.Eval(args[0],history),[STR],'msgKey') # must be string
        msg = ValidateValAndCast(self.qryRunner.Eval(args[1],history),[STR],'msg') # must be string
        self.__SendMessage(msgKey, msg, tid) #mark message for being sent
        val = BOL(True)
        return val
    
    def ExecScc(self, args:list, tid:int) -> BOL: 
        cmdId = ValidateValAndCast(args[0],[STR],'cmdId').GetVal() #everything must be string
        cmdStr = ValidateValAndCast(args[1],[STR, NON],'cmdStr').GetVal() #everything must be string
        #validate cmd form
        if(3 != len(cmdId) or (cmdId.upper() != cmdId)):
            raise EOQ_ERROR_INVALID_VALUE("cmdId must be 3 letter caps.")
        #see if the cmdId does already exist
        if(cmdId in self.cmdEvaluators):
            raise EOQ_ERROR_INVALID_VALUE("%s is reserved. Cannot set."%(cmdId))
        #try parsing the cmdStr in a command
        if(None == cmdStr): #remove the custom command
            if(cmdId in self.customCommands):
                del self.customCommands[cmdId]
        else: #register a new custom command
            try:
                cmd = self.cmdSerializer.DesCmd(cmdStr)
                #make sure the custom command is a compound command, because otherwise the history does not work
                if(not isinstance(cmd,Cmp)):
                    cmd = Cmp().Append(cmd)
                #check for recursions
                self.__PreventRecursion(cmd,cmdId)
                self.customCommands[cmdId] = cmd
            except Exception as e:
                raise EOQ_ERROR_INVALID_VALUE("Parsing cmdStr failed: %s"%(str(e)))
        val = BOL(True)
        return val

    def ExecVer(self, args:list, tid:int): 
        ''' Evaluates all constraints for a given target
        Returns:
        - [[constraint:Obj, isValid:BOL, exception:STR, changed:BOL],[...],...]
        '''
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[LST,Obj],'target')
        evaluationChangedEvents = []
        if(isinstance(target,Obj)):
            res = self.__VerifiyAllConstraintsForTarget(target)
            self.__GenerateVerificationChangeEvents(target,res,tid)
        else:
            if(not IsListOfObjects(target)):
                raise EOQ_ERROR_INVALID_VALUE("Requires Obj or LST<Obj>.")
            res = LST([])
            for t in target:
                r = self.__VerifiyAllConstraintsForTarget(t)
                self.__GenerateVerificationChangeEvents(t,r,tid)
                res.append(r)
        return res
    
    def __VerifiyAllConstraintsForTarget(self, target:Obj)->LST:
        allConstraints = self.__GetAllConstraintsRaw(target)
        # evaluate all applicable constraints
        verificationRecords = LST([])
        for c in allConstraints:
            valid,exception,changed = self.constraintVerifier.EvalConstraint(c, target)
            verificationRecords.append( LST([c,BOL(valid),STR(exception) if str==type(exception) else NON(),BOL(changed)]) )
        return verificationRecords
    
    def __GenerateVerificationChangeEvents(self, target:Obj, verificationRecords:LST,tid:int)->None:
        for r in verificationRecords:
            if(r[3].GetVal()==True): # is changed
                self.__AddLocalChange(tid, VecChange(target,r[0],r[1],r[2]),False,True)

    def ExecGac(self, args:list, tid:int): 
        ''' Retrieves all constraints applicable for a given target
        Returns:
        - [constraint1:Obj, ...]
        '''
        res = NON()
        history = self.__GetHistory(tid)
        target = ValidateVal(self.qryRunner.Eval(args[0],history),[LST,Obj],'target')
        if(isinstance(target,Obj)):
            res = LST(self.__GetAllConstraintsRaw(target))
        else:
            if(not IsListOfObjects(target)):
                raise EOQ_ERROR_INVALID_VALUE("Requires Obj or LST<Obj>.")
            res = LST([ LST(self.__GetAllConstraintsRaw(o)) for o in target])
        return res
        


    def __GetAllConstraintsRaw(self, target:Obj)->List[Obj]:
        '''Get all constraints applicable for the target.
        retrieve any inherited constraints, this is special for each concept type
        '''
        allConstraints = []
        concept = self.mdb.Read(target,STR(MXELEMENT.CONCEPT)).GetVal()
        # Currently no constraints applicable to M2 are defined
        # Constraints on M1 level are either direct of inherited from M1
        if(CONCEPTS.M1MODEL==concept):
            ownConstraints = self.mdb.Read(target,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += ownConstraints.GetVal()
            #check the m2model
            m2Model = self.mdb.Read(target,STR(M1MODEL.M2MODEL))
            inheritConstraints = self.mdb.Read(m2Model,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += inheritConstraints.GetVal()
        elif(CONCEPTS.M1OBJECT==concept):
            ownConstraints = self.mdb.Read(target,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += ownConstraints.GetVal()
            clazz = self.mdb.Read(target,STR(M1OBJECT.M2CLASS))
            classConstraints = self.mdb.Read(clazz,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += classConstraints.GetVal()
            #check all superclasses
            generalizations = self.mdb.Read(clazz,STR(M2CLASS.GENERALIZATIONS))
            for g in generalizations:
                clazz = self.mdb.Read(g,STR(M2INHERITANCE.SUPERCLASS))
                classConstraints = self.mdb.Read(clazz,STR(MXELEMENT.CONSTRAINTS))
                allConstraints += classConstraints.GetVal()
        elif(CONCEPTS.M1ATTRIBUTE==concept):
            ownConstraints = self.mdb.Read(target,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += ownConstraints.GetVal()
            m2Definition = self.mdb.Read(target,STR(M1ATTRIBUTE.M2ATTRIBUTE))
            inheritConstraints = self.mdb.Read(m2Definition,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += inheritConstraints.GetVal()
        elif(CONCEPTS.M1ASSOCIATION==concept):
            ownConstraints = self.mdb.Read(target,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += ownConstraints.GetVal()
            m2Definition = self.mdb.Read(target,STR(M1ASSOCIATION.M2ASSOCIATION))
            inheritConstraints = self.mdb.Read(m2Definition,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += inheritConstraints.GetVal()
        elif(CONCEPTS.M1COMPOSITION==concept):
            ownConstraints = self.mdb.Read(target,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += ownConstraints.GetVal()
            m2Definition = self.mdb.Read(target,STR(M1COMPOSITION.M2COMPOSITION))
            inheritConstraints = self.mdb.Read(m2Definition,STR(MXELEMENT.CONSTRAINTS))
            allConstraints += inheritConstraints.GetVal()
        return allConstraints
    
    def ExecCus(self, cmd:Cmd, args:list, tid:int) -> BOL: 
        ''' Executes a custom command registered externally
        '''
        transaction = self.__GetTransaction(tid)
        history = transaction.GetCurrentHistory()
        #resolve the arguments
        argVal = [self.qryRunner.Eval(a,history) for a in args]
        #add the argument values to the history of the command execution
        localHistory = transaction.GetNextHistory()
        for a in argVal:
            localHistory.AddValue(a,True)
        #execute the command
        val = self.ExecOnTransaction(cmd, tid, transaction.roc)
        return val
        
        
    
    #@Override
    def IsEventDesired(self, evt, sourceSessionId:str, sessionId:str)->bool:
        #extend decision based on the session
        if(sessionId):
            sessionIdPublic = self.__GetSessionIdPublic(sessionId)
            if(sessionIdPublic in self.sessions):
                session = self.sessions[sessionIdPublic]
                evtType = evt.a[0].GetVal()
                evtArgs = evt.a[1]
                #strategy = session.GetObservationStrategy(evtType)
                isObserved = session.IsEventObserved(evtType)
                isSuperobserved = session.IsEventSuperobserved(evtType) and self.accessController.IsAllowedToSuperobserve(session.user, evtType) #recheck if the rights have not changed
                if(not isObserved and not isSuperobserved):
                    return False #quick exit if event type is not desired
                #decide based on the event type what to do
                elif(EVT_TYPES.CRT == evtType):
                    elem = evtArgs[1] #the new object
                    createArgs = evtArgs[3]
                    if(sourceSessionId == sessionId):
                        return False #do not receive self induced changes
                    elif(isObserved):
                        #event selection for create depends on the create args
                        createArgObj = [e for e in createArgs if isinstance(e,Obj)]
                        #make sure that at least one object in the create args is observed
                        anyCreateArgObserved = any([session.IsElementObserved(e) for e in createArgObj])
                        allCreateArgsAllowed = all([self.mdb.IsAllowedToRead(e, NON(), session.user) for e in createArgObj]) 
                        return anyCreateArgObserved and allCreateArgsAllowed
                    elif(isSuperobserved):
                        return self.mdb.IsAllowedToRead(elem, NON(), session.user) 
                elif(EVT_TYPES.UPD == evtType):
                    elem = evtArgs[1] #the target
                    featureName = evtArgs[2]
                    value = evtArgs[3]
                    if(sourceSessionId == sessionId):
                        return False #do not receive self induced changes
                    elif(isObserved):
                        return session.IsElementObserved(elem) and\
                               self.mdb.IsAllowedToRead(elem, featureName, session.user) and\
                             (value.IsNone() or not isinstance(value,Obj) or self.mdb.IsAllowedToRead(value, NON(), session.user))
                    elif(isSuperobserved):
                        return self.mdb.IsAllowedToRead(elem, NON(), session.user) 
                elif(EVT_TYPES.DEL == evtType):
                    elem = evtArgs[1] #the target
                    if(sourceSessionId == sessionId):
                        return False #do not receive self induced changes
                    elif(isObserved):
                        return session.IsElementObserved(elem) and\
                               self.mdb.IsAllowedToRead(elem, NON(), session.user) 
                    elif(isSuperobserved):
                        return self.mdb.IsAllowedToRead(elem, NON(), session.user)
                elif(EVT_TYPES.VEC == evtType):
                    elem = evtArgs[0] #the target
                    const = evtArgs[1] #the constraint
                    if(sourceSessionId == sessionId):
                        return False #do not receive self induced changes
                    elif(isObserved):
                        return session.IsElementObserved(elem) and\
                               self.mdb.IsAllowedToRead(elem, NON(), session.user) and\
                               self.mdb.IsAllowedToRead(const, NON(), session.user)  
                    elif(isSuperobserved):
                        return self.mdb.IsAllowedToRead(elem, NON(), session.user) and\
                               self.mdb.IsAllowedToRead(const, NON(), session.user)   
                elif(EVT_TYPES.WAT == evtType):
                    if(sourceSessionId == sessionId): #each watcher only receives his private watch events
                        return True
                elif(EVT_TYPES.MSG == evtType):
                    if(sourceSessionId == sessionId):
                        return False #do not receive self induced messages
                    elif(isObserved):
                        msgKey = evtArgs[0].GetVal()
                        return session.IsMessageObserved(msgKey) 
                    elif(isSuperobserved):
                        return True
                    
        return False #without session ID and by default no events are received
        
    '''
    PRIVATE METHODS
    '''
    
    def __StartTransaction(self, tid:int, sessionId:str, readOnly:bool) -> Transaction:
        self.logger.Debug('Transaction %d started.'%(tid))
        transaction = Transaction(tid,readOnly,self.latestChangeId-1) #the current change is the latest-1.
        self.transactions[tid] = transaction
        session = self.__InitSession(transaction, sessionId)
        self.mdb.SetSession(session)
        return transaction
    
    def __EndTransaction(self, res:Cmd, tid:int) -> None:
        '''WARNING: __EndTransaction is not allowed to raise exceptions, because this will hold the domain forever
        '''
        try:
            transactionSucess = False #is set to true later
            #remove the current transaction from the stack
            transaction = self.transactions.pop(tid)
            #obtain the changes done during the transaction
            transactionChanges = transaction.changes
            #transactionSessionId = self.__MergeSessionId(transaction.session.sessionIdPublic, transaction.session.sessionIdSecret)
            if(res.cmd == CMD_TYPES.ERR):
                code = res.a[0]
                reason = res.a[1]
                self.logger.Warn('Transaction %d failed: %s: %s'%(tid,code,reason))
                nChanges = len(transactionChanges)
                if(0<nChanges):
                    self.logger.Warn('Transaction %d: Rolling back %d changes.'%(tid,nChanges))
                    self.__RollbackChanges(transactionChanges, transaction)
            else:
                #persist changes
                if(0<len(transactionChanges)):
                    self.__PersistChanges(transactionChanges,transaction.session)
                transactionSucess = True        
            #release the access controlled MDB
            self.mdb.UnsetSession()
            #POST PROCESSING: watched queries and events, but only if transaction succeeded
            if(transactionSucess):
                self.__UpdateWatchedQueriesAndCreateEvents()
                self.ReleasePostponedNotifications()
                self.__TriggerEventNotification()
            else: #transaction failed
                self.ClearPostponedNotifications();
            self.logger.Debug('Transaction %d ended.'%(tid))
        except Exception as e:
            #can not raise any error here, can only log a message.
            self.logger.Error("End transaction failed: %s"%(str(e)))
            
    def __UpdateWatchedQueriesAndCreateEvents(self)->None:
        '''Update any watched query for each watcher and release watch events.
        WARNING: this must be called after the main transaction is over, because this 
        function will toggle between all watching sessions.
        '''
        allWatchingSessions = self.queryWatcher.GetAllWatcherSessionIds()
        for s in allWatchingSessions:
            if(s in self.sessions):
                watcherSession = self.sessions[s]
                self.mdb.SetSession(watcherSession)
                watchChanges = self.queryWatcher.GetAllWatchChangesForSession(s)
                self.mdb.UnsetSession()
                watcherSessionId = watcherSession.sessionId
                for chg in watchChanges:
                    evt = self.__ChangeToEvent(chg)
                    self.PostponeObserverNotification([evt], watcherSessionId)
    
        
    def __GetTransaction(self, tid:int) -> Transaction:
        transaction = None
        try:
            transaction = self.transactions[tid]
        except KeyError:
            raise EOQ_ERROR_RUNTIME("Invalid transaction id %d"%tid)
        return transaction
    
    def __InitSession(self, transaction:Transaction, sessionId:str) -> Session:
        session = None
        if(None == sessionId):
            session = self.annonymousSession
        else:
            (sessionIdPublic, sessionIdSecret) = self.__SplitSessionId(sessionId)
            try:
                session = self.sessions[sessionIdPublic]
                if(session.sessionIdSecret != sessionIdSecret):
                    raise EOQ_ERROR_ACCESS_DENIED("Cannot use foreign session.")
            except KeyError:
                #if not existing, create a new one
                session = self.__CreateNewSession(sessionId, sessionIdPublic, sessionIdSecret)
        #set the current transaction to the session
        transaction.session = session
        return session
    
    def __SplitSessionId(self, sessionId:str)->Tuple[str,str]:
        if(SESSION_ID_LENGTH != len(sessionId)):
            raise EOQ_ERROR_INVALID_VALUE("Invalid session ID")
        sessionIdPublic = self.__GetSessionIdPublic(sessionId)
        sessionIdSecret = self.__GetSessionIdSecret(sessionId)
        return (sessionIdPublic,sessionIdSecret)
    
    def __GetSessionIdPublic(self, sessionId:str)->str:
        '''returns only the public part of the session ID
        '''
        return sessionId[SESSION_SECRET_LEN:]
    
    def __GetSessionIdSecret(self, sessionId:str)->str:
        '''returns only the secret part of the session ID
        '''
        return sessionId[:SESSION_SECRET_LEN]
    
    def __MergeSessionId(self, sessionIdPublic:str, sessionIdSecret:str)->str:
        ''' Composes public and secret part of a session id to the
        full session id
        ''' 
        return sessionIdSecret+sessionIdPublic
    
    def __CreateNewSession(self, sessionId:str, sessionIdPublic:str, sessionIdSecret:str, inheritSession:Session=None, isTemp:bool=False) -> Session:
        self.sessionCount += 1
        sessionNumber = self.sessionCount
        session = Session(sessionId, sessionIdPublic, sessionIdSecret, sessionNumber)
        session.isTemp = isTemp
        if(inheritSession):
            session.user = inheritSession.user #inherit the user and access rights
        self.sessions[sessionIdPublic] = session
        return session
        
    def __GetHistory(self, tid:int) -> History:
        return self.__GetTransaction(tid).GetCurrentHistory()

        
    def __AddLocalChange(self, tid:int, chg:Change, increasesChangeId:bool, isEvent:bool) -> None:
        '''Remembers every change done with in a transaction.
    
        Changes are stored to revert it in case of failures.
        Changes are only persisted if no failure occurs until the (compound) command is complete.
        
        '''
        transaction = self.__GetTransaction(tid)
        cid = self.latestChangeId + transaction.nRealChanges
        chg.timestamp = datetime.now()
        chg.cid = cid
        chg.tid = tid
        chg.user =  transaction.session.user
        chg.sessionNumber = transaction.session.sessionNumber
        #store the change in the transactions local change list
        self.__GetTransaction(tid).changes.append(chg)
        transaction.cid = cid #remember the changes in the transaction
        if(increasesChangeId):
            #calculate what the change ID will be if the change is persisted
            transaction.nRealChanges += 1
        if(isEvent):
            evt = self.__ChangeToEvent(chg)
            self.PostponeObserverNotification([evt], transaction.session.sessionId)
        
    def __PersistChanges(self, changes:list, session:Session) -> None:
        #notify observers on changes
        for chg in changes:
            #persist any model changes
            if(isinstance(chg,(CrtChange,UpdChange,DelChange))): #this are only the real changes
                #if the maximum of saved changes is exceeded begin deleting entries
                if(self.latestChangeId - self.earliestChangeId > self.config.cmdMaxChanges):
                    self.changes.pop(self.earliestChangeId)
                    self.earliestChangeId +=1
                #add each change of the current transaction to the change list
                self.changes[self.latestChangeId] = chg
                if(self.latestChangeId!=chg.cid):
                    #Should never go here
                    raise EOQ_ERROR_RUNTIME('Inconsistency in change detected list.')
                self.latestChangeId+=1
                #log the changes if logging is enabled
            else:
                pass #nothing to do, this will only result in an event
            self.logger.Debug('Change: %s'%(chg))
        
    def __ChangeToEvent(self,chg : Change) -> Evt:
        if(isinstance(chg,CrtChange)):
            return CrtEvt(U64(chg.cid),chg.target,chg.classId,chg.createArgs,InitValOrNon(chg.user,STR))
        elif(isinstance(chg,UpdChange)):
            return UpdEvt(U64(chg.cid),chg.target,chg.feature,chg.value,chg.position,InitValOrNon(chg.user,STR))
        elif(isinstance(chg,DelChange)):
            return DelEvt(U64(chg.cid),chg.target,InitValOrNon(chg.user,STR))
        elif(isinstance(chg,MsgChange)):
            return MsgEvt(chg.msgKey,chg.msg)
        elif(isinstance(chg,VecChange)):
            return VecEvt(chg.target,chg.constraint,chg.isValid,chg.error)
        elif(isinstance(chg,WatChange)):
            return WatEvt(chg.wid,chg.result,chg.error)
        else: #should never go here
            raise EOQ_ERROR_RUNTIME('Experienced unexpected change type: %s'%(type(chg).__name__))
            
    def __ChangeToChangeRecord(self, chg:Change) -> LST:
        if(isinstance(chg,CrtChange)):
            return LST([STR(EVT_TYPES.CRT),U64(chg.cid),chg.target,chg.classId,chg.createArgs,InitValOrNon(chg.user,STR)])
        elif(isinstance(chg,UpdChange)):
            return LST([STR(EVT_TYPES.UPD),U64(chg.cid),chg.target,chg.feature,chg.value,chg.position,InitValOrNon(chg.user,STR)])
        elif(isinstance(chg,DelChange)):
            return LST([STR(EVT_TYPES.DEL),U64(chg.cid),chg.target,InitValOrNon(chg.user,STR)])
        else: #should never go here
            raise EOQ_ERROR_RUNTIME('Experienced unexpected change type: %s'%(type(chg).__name__))
        
    def __RollbackChanges(self, changes:list, transaction:Transaction)->None:
        #revert local changes for multicommand transactions
        try:
            for chg in reversed(changes):
                self.logger.Warn("ROLLBACK: %s "%(chg))
                if(isinstance(chg,CrtChange)):
                    self.__RollbackCrt(chg)
                elif(isinstance(chg,UpdChange)):
                    self.__RollbackUpd(chg)
                elif(isinstance(chg,DelChange)):
                    self.__RollbackDel(chg)
                elif(isinstance(chg,ObsChange)):
                    self.__RollbackObs(chg,transaction)
                elif(isinstance(chg,UbsChange)):
                    self.__RollbackUbs(chg,transaction)
                else: #should never go here
                    pass # nothing to do for other changes
        except Exception as e:
            self.logger.Error("Error during rollback: %s"%(str(e)))  
            if self.config.printUnexpectedExceptionTraces:
                print("Unexpected exception trace (__RollbackChanges):") 
                traceback.print_exc()
            
    def __RollbackCrt(self, chg:CrtChange):
        self.mdb.Delete(chg.target,False,not chg.postModError) #use unprotected mdb to make sure that rollback does not fail
            
    def __RollbackUpd(self, chg:UpdChange)->None:
        position = chg.position
        #reset the old value
        if(chg.value.IsNone()):
            if(position >= 0):
                position = -(position+I64(2)) #if a value was deleted, it must be re-added here
        self.mdb.Update(chg.target,chg.feature,chg.oldValue,position,False,not chg.postModError)
        #rebuild the old values containment if it was existing before
        if(chg.oldOwner):
            oldFeatureName = self.mdb.Read(chg.oldFeature,STR(M2COMPOSITION.NAME),NON(),False,not chg.postModError)
            self.mdb.Update(chg.oldOwner,oldFeatureName,chg.value,-(chg.oldIndex+I64(2)),False,not chg.postModError)
            
    def __RollbackDel(self, chg:DelChange)->None:
        (elem,pmerr) = self.mdb.Create(chg.classId,chg.createArgs,chg.target,chg.recoveryArgs) #restore the element with its old ID
                
    def __RollbackObs(self, chg:ObsChange, transaction:Transaction)->None:
        evtType = chg.evtType
        evtKey = chg.evtKey
        history = None
        session = transaction.session
        self.__Ubs(evtType, evtKey, history, session)
        
    def __RollbackUbs(self, chg:ObsChange, transaction:Transaction)->None:
        evtType = chg.evtType
        evtKey = chg.evtKey
        history = None
        if(chg.affectedSessionPublicId in self.sessions):
            affectedSession = self.sessions[chg.affectedSessionPublicId]
            self.__Obs(evtType, evtKey, history, affectedSession)
    
    def __PreventRecursion(self,cmd,cmdId):
        if(cmd.cmd == cmdId):
            raise EOQ_ERROR_INVALID_VALUE('No recursion allowed.')
        elif(cmd.cmd == CMD_TYPES.CMP): #compound command?
            for c in cmd.a:
                self.__PreventRecursion(c, cmdId)
                
    def __SendMessage(self, msgKey:STR, msg:STR, tid:int)->None:
        '''Sends a message to observers. 
        The message is not directly send but put to the event stack and send 
        when the transaction is over
        '''
        if(self.config.cmdMaxMsgKeyLength<len(msgKey)):
            raise EOQ_ERROR_INVALID_VALUE('Message key has more than %d chars'%(self.config.cmdMaxMsgKeyLength))
        if(self.config.cmdMaxMsgLength<len(msg)):
            raise EOQ_ERROR_INVALID_VALUE('Message has more than %d chars: %s'%(self.config.cmdMaxMsgLength,msg))
        change = MsgChange(msgKey, msg)
        self.__AddLocalChange(tid,change,False,True)
        
    def __UnobserveDeletedElementForAllSessions(self, tid:int, deletedElement:Obj)->None:
        '''Creates UbsChange events for all sessions
        '''
        for session in self.sessions.values():
            isChanged = session.UnobserveElement(deletedElement)
            if(isChanged):
                change = UbsChange(EVT_TYPES.ELM,deletedElement,session.sessionIdPublic)
                self.__AddLocalChange(tid,change,False,False)
        
    def __DeprecatedWarning(self, message:str):
        self.logger.Warn("%s is deprecated. Consider replacement."%(message))
        
        
    ### EVENT NOTIFICATION ###
    def __TriggerEventNotification(self):
        '''Triggers that queued and released are sent to their observers. 
        This either happens synchronous or asynchronous
        '''
        if(self.config.evtNotifyAsync):
            self.notifyObserverSignal.set()
        else:
            self.SendQueuedAndReleasedNotifications()
    
    def __EventNotificationLoop(self)->None:
        while(self.shallRun):
            hasNewEvents = self.notifyObserverSignal.wait(timeout=self.config.threadLoopTimeout)
            #hasNewEvents = self.evtSemaphore.acquire(timeout=self.config.threadLoopTimeout)
            if(hasNewEvents):
                self.SendQueuedAndReleasedNotifications()
                self.notifyObserverSignal.clear()
        #finally make sure no events get lost
        self.SendQueuedAndReleasedNotifications()
        



'''
 Bjoern Annighoefer 2022
'''


from .eoq3Lexer import eoq3Lexer
from .eoq3Listener import eoq3Listener
from .eoq3Parser import eoq3Parser

from ..serializer import Serializer, RegisterSerializer
 
from ...frame import Frm
from ...value import VAL, BOL, U32, U64, I32, I64, F32, F64, STR, DAT, NON, QRY, LST, ValCompare
from ...query import Qry, Seg
from ...command import Cmd, Cmp, CmdFactory, GetCmdInfo, PrmCmd
from ...error import EOQ_ERROR_INVALID_VALUE, EOQ_ERROR_INVALID_TYPE, EOQ_ERROR_RUNTIME

from antlr4 import InputStream, CommonTokenStream, ParseTreeWalker
from antlr4.error.ErrorListener import ErrorListener

import re

from typing import List, Any


#define a list of symbols for the textual representation
QRY_SYMBOLS = {
    'OBJ' : '#',
    'ONI' : '`', 
    'HIS' : '$', 
    'PTH' : '/', 
    'CLS' : '!', 
    'INO' : '?',
    'MET' : '@',
    'NOT' : '&NOT',
    'TRM' : '&TRM',
    'TRY' : '&TRY',
    'QRF' : '&QRF',
    'STF' : '&STF',
    'SLF' : '&SLF',
    'IDX' : ':', 
    'SEL' : '{', 
    'ARR' : '&ARR', 
    'ZIP' : '&ZIP',
    'ANY' : '&ANY',
    'ALL' : '&ALL', 
    'EQU' : '=',
    'EQA' : '&EQA',
    'NEQ' : '~', 
    'LES' : '<', 
    'GRE' : '>',
    'RGX' : '&RGX',
    'ADD' : '&ADD', 
    'SUB' : '&SUB', 
    'MUL' : '&MUL', 
    'DIV' : '&DIV', 
    'ORR' : '&ORR', 
    'XOR' : '&XOR', 
    'AND' : '&AND', 
    'NAD' : '&NAD', 
    'CSP' : '&CSP', 
    'ITS' : '&ITS', 
    'DIF' : '\\',
    'UNI' : '&UNI', 
    'CON' : '|'
}  
#free symbols: `´
#reserved symbols: *{}()[]'"+-.,

# special commands for handling single and no command cmp commands
CMD_TYPE_PDD = 'PDD' #pseudo padding command type
class Pdd(PrmCmd):
    def __init__(self):
        super().__init__(CMD_TYPE_PDD, [],True,None)

# create a look up table for single char query symbols
QRY_SYMBOLS_LOT = {} #contains query char->3 letter encoding
for k,v in QRY_SYMBOLS.items(): 
    if('&' != v[0]):
        QRY_SYMBOLS_LOT[v] = k

STR_SIMPLE_1 = "[A-Za-z0-9+\\-*_]*[A-Za-z+\\-*._]" # copied from .g4 file
STR_SIMPLE_2 = "[A-Za-z0-9+\\-*_][A-Za-z*._][A-Za-z0-9+\\-*_]*" # copied from .g4 file
STR_SIMPLE_3 = "[ABCEFGHIJKLMNOPQRSTUVWXYZaceghjkmopqrstvwxz+\\-*_]+[A-Za-z0-9+\\-*._]*" # copied from .g4 file
STR_SIMPLE = "(%s)|(%s)|(%s)"%(STR_SIMPLE_1,STR_SIMPLE_2,STR_SIMPLE_3)

STR_SIMPLE_RGX = re.compile(STR_SIMPLE)

CMD_INDENT = '    ' #one level of indent are 4 spaces


def CharAllowed(c,whitelistdict):
    try:
        return whitelistdict[c]
    except KeyError:
        return False

class Eoq3TextAntlrErrorListener(ErrorListener):
    #@override
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise EOQ_ERROR_INVALID_VALUE("line %d:%d %s"%(line,column,msg))

class Eoq3TextAntlrListener(eoq3Listener):
    def __init__(self):
        super().__init__()
        self.stack:List[Any] = []  #initialize the stack
        self.indent:List[int] = [] #initialize the indent counter
        self.stackTop = 0
        self.indentLevel = -1  # initialize to -1 to make sure level 0 is initialized when entered

    def __HandleIndent(self,currentIndent:int,ctx):
        ''' handle indent for forming compound commands
        '''
        if(self.indentLevel == currentIndent): #no change in indent
            self.indent[currentIndent] += 1
        elif(self.indentLevel+1 == currentIndent):# indent
            self.indentLevel = currentIndent
            self.indent.append(1)
        elif(self.indentLevel > currentIndent):# dedent
            self.__Dedent(currentIndent)
            self.indent[currentIndent] += 1
        else: 
            raise EOQ_ERROR_INVALID_VALUE('Incorrect indent at %s'%(ctx.getText()))
        
    def __Dedent(self,currentIndent:int):
        while(self.indentLevel > currentIndent):
            #get all indented cmds
            cmds = self.__RemovePddArgs(self.__PopN(self.indent[self.indentLevel]))
            father = self.__Pop()
            if(isinstance(father,Cmp)): #explicit cmp command
                for c in cmds:
                    father.Append(c)
                self.__Push(father)
            else: #implicit cmp command
                self.__Push(father) #this was not a cmp, put it pack
                newFather = Cmp(cmds)
                self.__Push(newFather)
                self.indent[self.indentLevel-1] += 1
            self.indentLevel -= 1
            self.indent.pop()
         
    def GetResult(self):
        return self.__Pop()
        
    def exitCmds(self, ctx):
        #make sure commands are dedented to zero again
        self.__Dedent(0)
        #prepare return value, which is either a single command or a compound
        nCmds = self.indent[0]
        if(1==nCmds):
            pass #do nothing since the cmd is already on the stack
        else:
            #create a compound command
            cmds = self.__PopN(nCmds)
            cmd = Cmp(self.__RemovePddArgs(cmds)) #remove padding commands
            self.__Push(cmd)
         
    def exitCmd(self, ctx):
        cmdIdStr = ctx.CMD_ID().getText()
        #mute
        mute = False
        cmdId = cmdIdStr[0:3]
        if('-' == cmdIdStr[0]):
            mute = True
            cmdId = cmdIdStr[1:4] 
        #res name: handle first, because res name is last on stack
        resName = None
        if ctx.cmd_res_name():
            resName = self.__Pop().GetVal()
        #args
        nArgs = len(ctx.cmd_arg())
        args = self.__PopN(nArgs)
        # indent handling
        currentIndent = len(ctx.CMD_INDENT())
        self.__HandleIndent(currentIndent, ctx)
        #create command
        cmd = None
        if(CMD_TYPE_PDD == cmdId):
            cmd = Pdd()
        else:
            cmd = CmdFactory(cmdId,args,mute,resName)
        #push parsed cmd
        self.__Push(cmd)
             
    # handle primitives  
    def exitVal_bol(self,ctx):
        strVal = ctx.getText()
        val = BOL(1 == int(strVal[1])) #second char can be '0' or '1'
        self.__Push(val)
         
    def exitVal_u32(self,ctx):
        strVal = ctx.getText()
        val = U32(int(strVal[1:])) #remove the first char
        self.__Push(val)
         
    def exitVal_u64(self,ctx):
        strVal = ctx.getText()
        val = U64(int(strVal[1:])) #remove the first char
        self.__Push(val)
         
    def exitVal_i32(self,ctx):
        strVal = ctx.getText()
        val = I32(int(strVal[1:])) #remove the first char
        self.__Push(val)
         
    def exitVal_i64(self,ctx):
        strVal = ctx.getText()
        val = I64(int(strVal[1:])) #remove the first char
        self.__Push(val)
         
    def exitVal_f32(self,ctx):
        strVal = ctx.getText()
        val = F32(float(strVal[1:])) #remove the first char
        self.__Push(val)
     
    def exitVal_f64(self,ctx):
        strVal = ctx.getText()
        val = F64(float(strVal[1:])) #remove the first char
        self.__Push(val)
     
    def exitVal_str_simple(self, ctx):
        val = STR(ctx.getText())
        self.__Push(val)
         
    def exitVal_str_quote(self, ctx):
        quotedStr = ctx.getText()
        val = STR(quotedStr[1:-1].replace("\\\'","'").replace("\\\\","\\")) #the surrounding quotes and replace escape characters
        self.__Push(val)
         
    def exitVal_dat(self,ctx):
        #initialize 0 date
        Y = 0; M = 0; D=0 #year, month, day
        h = 0; m = 0; s= 0 #hour, minute, second
        u = 0.0 #microseconds
        z = 0 #timezone
        #time is mandatory
        datetimeStr = ctx.getText()
        dStart = datetimeStr.find('D')
        tStart = datetimeStr.find('T')
        uStart = datetimeStr.find('U')
        zStart = datetimeStr.find('Z')
        if(0<=dStart): #optional
            dateSegs = datetimeStr[dStart+1:dStart+11].split('-') #expecting 0000-00-00
            Y = int(dateSegs[0]); M = int(dateSegs[1]); D = int(dateSegs[2])
        if(0<=tStart): #mandatory
            timeSegs = datetimeStr[tStart+1:tStart+9].split('.') #expecting 00.00.00
            h = int(timeSegs[0]); m = int(timeSegs[1]); s = int(timeSegs[2])
        if(0<=uStart):
            if(0<=zStart):
                u = float(datetimeStr[uStart+1:zStart])
            else:
                u = float(datetimeStr[uStart+1:])
        if(0<=zStart):
            z = float(datetimeStr[zStart+1:])
        val = DAT(Y,M,D,h,m,s,u,z) 
        self.__Push(val)
         
    def exitVal_non(self,ctx):
        val = NON()
        self.__Push(val)
         
    def exitVal_lst(self,ctx):
        nElems = len(ctx.val()) #determine the length of the list
        val = self.__PopN(nElems) #remove those elements from the list ...
        self.__Push(LST(val))   #... and push it back as a nested list.
         
    def exitVal_qry(self,ctx):
        nSegs = len(ctx.qry_seg()) #determine the length of the list
        try:
            segs = self.__PopN(nSegs) #remove those elements from the list ...
        except Exception as e: #TODO: remove this if problems with special constraint 
            print("TXTSER FAILED: %s"%(str(ctx.getText())))
            raise e
        qry = Qry(segs)
        self.__Push(qry)   #... and push it back as a nested list.
         
    def exitQry_seg(self,ctx):
        nArgs = len(ctx.qry_arg()) #determine the length of the list
        try:
            args = self.__PopN(nArgs) #remove those elements from the list ...
        except Exception as e: #TODO: remove this if problems with special constraint 
            print("TXTSER FAILED: %s"%(str(ctx.getText())))
            print("TXTSER STACK: %s"%(str(self.stack)))
            raise e
        segType = self.__Pop() #the segment type was stored before the arguments
        seg = Seg(segType,args)
        self.__Push(seg)   #... and push it back as a nested list.
         
    def exitQry_sym_short(self,ctx):
        shortSym = ctx.getText() #a single char
        try:
            segType = QRY_SYMBOLS_LOT[shortSym]
            self.__Push(segType)
        except KeyError:
            raise EOQ_ERROR_INVALID_VALUE('Unknown query segment: %s'%(shortSym))
         
    def exitQry_sym_full(self,ctx):
        full = ctx.getText() # a three letter char starting with &
        segType = full[1:]
        if(segType in QRY_SYMBOLS): #this just tests if the symbol is known
            self.__Push(segType)
        else:
            raise EOQ_ERROR_INVALID_VALUE('Unknown query segment: %s'%(full))
     
    def __Pop(self):
        if(self.stackTop-1 >= 0):
            self.stackTop -= 1
            return self.stack.pop()
        else:
            raise EOQ_ERROR_RUNTIME('Can not pop element from stack. Stack is empty')
         
    def __PopN(self,n=1)->List[Any]:
        if(self.stackTop-n >= 0):
            self.stackTop -= n
            return [self.stack.pop(i) for i in range(-n,0)]
        else:
            raise EOQ_ERROR_RUNTIME('Can not pop %d elements from stack. Stack has only %d elements.'%(n,self.stackTop))
         
    def __Push(self,v):
        self.stack.append(v)
        self.stackTop += 1

    def __RemovePddArgs(self, args:List[Any])->List[Any]:
        '''Remove padding commands from the list of arguments
        Returns the list of arguments without padding commands
        '''
        return [a for a in args if not isinstance(a,Pdd)]


class TextSerializer(Serializer):
    ''' A text serializer based on Antlr 4
    Args:
        skipDefaultValues:bool (default=True): if true the command
            serializer will leave out arguments that have default values if
            at the end of the command
    
    '''
    def __init__(self,skipDefaultValues:bool=True):
        #settings
        self.skipDefaultValues = skipDefaultValues
        #initialization
        self.errorListener = Eoq3TextAntlrErrorListener()

    def Name(self):
        '''Returns a three letter identifier for this kind of serialization 
        '''
        return 'TXT'
    
    ### SERIALIZATION

    # @Overrides
    def SerFrm(self, frm:Frm)->str:
        '''Textual frames have a fixed length, which simplifies deserialization'''
        sid = '' if None==frm.sid else frm.sid
        return "AMI|%8s|%3s|%0.8x|%36s|%3s|%1d\n%s"%(frm.ver.ljust(8, ' '),frm.typ,frm.uid,sid,frm.ser,frm.roc,frm.dat)
    
    # @Override
    def SerCmd(self, cmd:Cmd)->str:
        return self.__SerCmdRaw(cmd,0,False)
    
    def __SerCmdRaw(self, cmd:Cmd, level:int, forceCmp:bool)->str:
        indentStr = ''.join([CMD_INDENT for i in range(level-1)]) #-1 = no indent for first level
        if(isinstance(cmd,Cmp)):
            nCmd = len(cmd.a) #number of subcommands
            prefix = ''
            cmdStrs = []
            isAfterCmp = True #the first command is always after a cmp
            # serialize all subcommands
            for c in cmd.a: 
                cmdStrs.append(self.__SerCmdRaw(c,level+1,isAfterCmp))
                isAfterCmp = isinstance(c,Cmp)
            # padding for 0 and 1 command cmps
            if(1 == nCmd):  # special case, because a cmp with only one subcommand cannot be distinguished from a single command
                # pad by one PDD command
                cmdStrs.append(self.__SerCmdRaw(Pdd(),level+1,False))
            elif(0 == nCmd): #special case, because an empty cmp results in no serialization
                # pad by two PDD command
                cmdStrs.append(self.__SerCmdRaw(Pdd(), level + 1, False))
                cmdStrs.append(self.__SerCmdRaw(Pdd(), level + 1, False))
            # serialize the cmp itself if necessary, i.e. mute or result name is set, but not on the first level
            if(level>0 and (forceCmp or cmd.m or None != cmd.r)):# the first cmp is implicit if no mute or result name is specified
                prefix = indentStr+self.__SerCmdMute(cmd)+cmd.cmd+self.__SerCmdResName(cmd)+'\n'
            return prefix + ('\n'.join(cmdStrs))
        else:
            mandArgs = []
            cmdInfo = GetCmdInfo(cmd.cmd)
            if(None != cmdInfo and self.skipDefaultValues):
                #if command is known, show only optional arguments with a value different from the default.
                #find out how many optional args have values identical to the default value, i.e find the first optional argument with an non-default value
                lastNonDefaultOptArg = 0
                for i in range(cmdInfo.optArgs):
                    if(not ValCompare(cmd.a[cmdInfo.minArgs+i],cmdInfo.optDefVals[i])):
                        lastNonDefaultOptArg = i+1
                #strip the optional arguments as long as they match their default value
                mandArgs = cmd.a[0:cmdInfo.minArgs+lastNonDefaultOptArg]
            else:
                mandArgs = cmd.a
            argStrs = [' '+self.SerVal(a) for a in mandArgs]
            return indentStr+self.__SerCmdMute(cmd)+cmd.cmd+(''.join(argStrs))+self.__SerCmdResName(cmd)
        
    def __SerCmdMute(self, cmd:Cmd):
        return ('-' if cmd.m else '')
    
    def __SerCmdResName(self, cmd:Cmd):
        return ((' -> $'+self._SerStr(cmd.r)) if cmd.r else '')
        
    # @Override    
    def SerQry(self, qry:QRY)->str:
        segStrs = [self.SerSeg(s) for s in qry.v]
        return '('+''.join(segStrs)+')'
    
    # @Override
    def SerSeg(self, seg:Seg)->str:
        argStrs = [self.SerVal(a) for a in seg.v]
        return QRY_SYMBOLS[seg.seg]+';'.join(argStrs)
    
    # @Override    
    def SerVal(self, val:VAL)->str:
        if(isinstance(val,NON)): 
            return 'n0'
        elif(isinstance(val,BOL)): 
            return 'b1' if val.v else 'b0'
        elif(isinstance(val,U32)): 
            return 'u%d'%(val.v)
        elif(isinstance(val,U64)): 
            return 'y%d'%(val.v)
        elif(isinstance(val,I32)): 
            return 'i%d'%(val.v)
        elif(isinstance(val,I64)): 
            return 'l%d'%(val.v)
        elif(isinstance(val,F32)): 
            return ('f%f'%(val.v)).rstrip('0').rstrip('.') #remove trailing 0 and .
        elif(isinstance(val,F64)): 
            return ('d%f'%(val.v)).rstrip('0').rstrip('.') #remove trailing 0 and .
        elif(isinstance(val,STR)): 
            return self._SerStr(val.v)
        elif(isinstance(val,DAT)): 
            return 'D%04d-%02d-%02dT%02d.%02d.%02dU%010.7fZ%+06.2f'%(val.year(),val.month(),val.day(),val.hour(),val.minute(),val.second(),val.microsec(),val.utcoff())
        elif(isinstance(val,QRY)):
            return self.SerQry(val)
        elif(isinstance(val,LST)): #list must be the after str, because str is also a list
            valStrs = [self.SerVal(v) for v in val.v] #cannot use Val() here, because that decomposes the full list
            return '['+','.join(valStrs)+']'
        else:
            raise EOQ_ERROR_INVALID_TYPE("Unknown value type %s."%(type(val).__name__))
    
    # @Override
    def _SerStr(self,val:str)->str:
        n = len(val)
        if(0 == n):
            return self._SerStrComplex(val)
        elif(STR_SIMPLE_RGX.fullmatch(val)):
            return val
        else:
            return self._SerStrComplex(val)
        
    def _SerStrComplex(self, val:str):
            return "'"+val.replace("\\", "\\\\").replace("'", "\\'")+"'"  
        
        
    #DESERIALIZATION   
    def __Parse(self, text):
        lexerInput = InputStream(text)
        lexer = eoq3Lexer(lexerInput) #input stream is provided later
        lexer.removeErrorListeners()
        lexer.addErrorListener(self.errorListener)
        stream = CommonTokenStream(lexer)
        parser = eoq3Parser(stream)
        parser.removeErrorListeners()
        parser.addErrorListener(self.errorListener)
        return parser

    def __Construct(self, tree):
        constructListener = Eoq3TextAntlrListener()
        walker = ParseTreeWalker()
        walker.walk(constructListener, tree)
        res = constructListener.GetResult()
        return res
    
    # @Override
    def DesFrm(self,data:str)->Frm:
        '''the frame length is not checked'''
        #"AMI|%8s|%3s|%8x|%36s|%3s\n%s"(frm.ver,frm.typ,frm.uid,frm.sid,frm.ser,frm.dat)
        #Example:
        #AMI|33.22.11|CMD|3ade68b1|95103b3e-293c-48d6-9c07-2f97e4ab3690|TXT
        #    4        13  17       26                                   63
        ver = data[4:12].split(' ')[0] #only keep the part until the first blank
        typ = data[13:16]
        uid = int(data[17:25],base=16)
        sid = data[26:62]
        #sids that do not fill the first char are assumed as none
        sid = None if sid[0] == ' ' else sid
        ser = data[63:66]
        roc = bool(int(data[67:68]))
        dat = data[69:]
        return Frm(typ, uid, ser, dat, sid, roc, ver)
    
    # @Override
    def DesCmd(self,data:str)->Cmd:
        parser = self.__Parse(data)
        tree = parser.cmds()
        cmd = self.__Construct(tree)
        return cmd
    
    # @Override
    def DesQry(self, data:str)->Qry:
        parser = self.__Parse(data)
        tree = parser.val_qry()
        qry = self.__Construct(tree)
        return qry
    
    # @Override
    def DesSeg(self, data:str)->Seg:
        parser = self.__Parse(data)
        tree = parser.qry_seg()
        seg = self.__Construct(tree)
        return seg
    
    # @Override
    def DesVal(self, data:str):
        parser = self.__Parse(data)
        tree = parser.val()
        val = self.__Construct(tree)
        return val
    
### make this serializer known globally 
RegisterSerializer(TextSerializer)


'''
 deleter.py recursive delete function for a model tree based on concepts. 
 
 This file is generated form: deleter.py.mako on 2024-07-10 09:59:28.377607 by calling E:/1_ILS/Projects/2018_EOQ/Groups/PYEOQ/pyeoq/eoq3conceptsgen/generatefromconceptscli.py -i deleter.py.mako -o ../deleter.py.
 
 Bjoern Annighoefer 2023
'''

from ..concepts import CONCEPTS, MXMDB ,MXELEMENT ,MXCONSTRAINT ,M2PRIMITIVES ,M2PACKAGE ,M2MODEL ,M2ENUM ,M2OPTIONOFENUM ,M2CLASS ,M2ATTRIBUTE ,M2ASSOCIATION ,M2COMPOSITION ,M2INHERITANCE ,M1MODEL ,M1OBJECT ,M1FEATURE ,M1ATTRIBUTE ,M1ASSOCIATION ,M1COMPOSITION
from ..value import VAL, STR, NON, LST, I64
from ..query import Obj
from ..error import EOQ_ERROR_RUNTIME
from .util import IsList

from .conceptwalker import ConceptWalker

from typing import Tuple, Any, Dict, List

class DeleterRun:
    '''This class is used to maintain global information in one delete call
    '''
    def __init__(self):
        self.deletedElements:Dict[Obj,bool] = {} # the delete path in the model is not circle free and cannot be circle free, so need this to trac deleted elements
        self.deleteChain:List[Obj] = [] #the order in which objects are deleted at the end
        self.result:Tuple[STR,LST,LST] = (NON(),LST([]),LST([])) #the return value of the last delete command
 

class Deleter(ConceptWalker):
    def __init__(self,cmdrunner,logger):
        super().__init__(cmdrunner)
        self.logger = logger
        
    def DeleteAuto(self, target, tid)->Tuple[STR,LST,LST]:
        '''Delete target and its delete path properties.
        '''
        deleterRun = DeleterRun()
        # use walk to collect objects to delete and order, but do not delete, because that would change the struture during walk
        # moreover unset any attribute to enable deletion afterwards
        self.Walk(tid,target,NON(),1,deleterRun)
        #when walk is finished, elements can be deleted
        for e in deleterRun.deleteChain:
            deleterRun.result = self.cmdrunner._MdbDelete(tid,e)
        return deleterRun.result
        
    def DeleteFull(self, target:Obj, tid:int)->Tuple[STR,LST,LST]:
        '''Delete target and its delete path properties recursively
        '''
        deleterRun = DeleterRun()
        # use walk to collect objects to delete and order, but do not delete, because that would change the struture during walk
        # moreover unset any attribute to enable deletion afterwards
        self.Walk(tid,target,NON(),-1,deleterRun)
        #when walk is finished, elements can be deleted
        for e in deleterRun.deleteChain:
            deleterRun.result = self.cmdrunner._MdbDelete(tid,e)
        return deleterRun.result
    
    def __MdbDelete(self, tid:int, target:Obj, deleterRun:DeleterRun)->None:
        if(target not in deleterRun.deletedElements):
            deleterRun.result = self.cmdrunner._MdbDelete(tid,target)
            deleterRun.deletedElements[target] = True
            
    def __MarkForDeletion(self, tid:int, target:Obj, deleterRun:DeleterRun)->None:
        if(target not in deleterRun.deletedElements):
            deleterRun.deleteChain.append(target)
            deleterRun.deletedElements[target] = True
    
    def __MdbUpdate(self, tid:int, target:Obj, featureName:STR, value:VAL, pos:I64):
            self.cmdrunner._MdbUpdate(tid,target,featureName,value,pos)

    #@override
    def _OnExitMxMdb(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        # M2MODELS
        elements = self.cmdrunner._MdbRead(tid,target,STR(MXMDB.M2MODELS))
        nElements = len(elements)
        if(0<nElements):
            for i in range(nElements-1,-1,-1):
                e = elements[i]
                self.__MarkForDeletion(tid,e,data)
        # M1MODELS
        elements = self.cmdrunner._MdbRead(tid,target,STR(MXMDB.M1MODELS))
        nElements = len(elements)
        if(0<nElements):
            for i in range(nElements-1,-1,-1):
                e = elements[i]
                self.__MarkForDeletion(tid,e,data)
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitMxElement(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        # DOCUMENTATION
        e = self.cmdrunner._MdbRead(tid,target,STR(MXELEMENT.DOCUMENTATION))
        if(not e.IsNone()): i=0; self.__MdbUpdate(tid,target,STR(MXELEMENT.DOCUMENTATION),NON(),i)
        # OWNER
        e = self.cmdrunner._MdbRead(tid,target,STR(MXELEMENT.OWNER))
        if(not e.IsNone()): i=0; self.__MdbUpdate(tid,target,STR(MXELEMENT.OWNER),NON(),i)
        # GROUP
        e = self.cmdrunner._MdbRead(tid,target,STR(MXELEMENT.GROUP))
        if(not e.IsNone()): i=0; self.__MdbUpdate(tid,target,STR(MXELEMENT.GROUP),NON(),i)
        # PERMISSIONS
        elements = self.cmdrunner._MdbRead(tid,target,STR(MXELEMENT.PERMISSIONS))
        nElements = len(elements)
        if(0<nElements):
            for i in range(nElements-1,-1,-1):
                e = elements[i]
                self.__MdbUpdate(tid,target,STR(MXELEMENT.PERMISSIONS),NON(),i)
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitMxConstraint(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Primitives(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Package(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Enum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2OptionOfEnum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Class(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        # MYDSTASSOCIATIONS
        elements = self.cmdrunner._MdbRead(tid,target,STR(M2CLASS.MYDSTASSOCIATIONS))
        nElements = len(elements)
        if(0<nElements):
            for i in range(nElements-1,-1,-1):
                e = elements[i]
                self.__MarkForDeletion(tid,e,data)
        # MYCHILDCOMPOSITIONS
        elements = self.cmdrunner._MdbRead(tid,target,STR(M2CLASS.MYCHILDCOMPOSITIONS))
        nElements = len(elements)
        if(0<nElements):
            for i in range(nElements-1,-1,-1):
                e = elements[i]
                self.__MarkForDeletion(tid,e,data)
        # MYSPECIALIZATIONS
        elements = self.cmdrunner._MdbRead(tid,target,STR(M2CLASS.MYSPECIALIZATIONS))
        nElements = len(elements)
        if(0<nElements):
            for i in range(nElements-1,-1,-1):
                e = elements[i]
                self.__MarkForDeletion(tid,e,data)
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM2Inheritance(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM1Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM1Object(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        # DSTASSOCIATIONS
        elements = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.DSTASSOCIATIONS))
        nElements = len(elements)
        if(0<nElements):
            for i in range(nElements-1,-1,-1):
                e = elements[i]
                self.__MarkForDeletion(tid,e,data)
        # CHILDCOMPOSITION
        e = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.CHILDCOMPOSITION))
        if(not e.IsNone()): i=0; self.__MarkForDeletion(tid,e,data)
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM1Feature(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM1Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM1Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

    #@override
    def _OnExitM1Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        if(successor!=target): #prevent the delete call for super
            self.__MarkForDeletion(tid,target,data)

